# Backward compat.
from ..estimation.alignment import AlignmentBase
from ..estimation.cor import CenterOfRotation, CenterOfRotationAdaptiveSearch, CenterOfRotationGrowingWindow, CenterOfRotationSlidingWindow
from ..estimation.translation import DetectorTranslationAlongBeam
from ..estimation.focus import CameraFocus
from ..estimation.tilt import CameraTilt




"""
The following de-striping method is adapted from the pore3d software.

:Organization: Elettra - Sincrotrone Trieste S.C.p.A.

:Version: 2013.05.01

References
----------
[1] F. Brun, A. Accardo, G. Kourousias, D. Dreossi, R. Pugliese. Effective implementation
of ring artifacts removal filters for synchrotron radiation microtomographic images. Proc.
of the 8th International Symposium on Image and Signal Processing (ISPA), pp. 672-676, Sept. 4-6,
Trieste (Italy), 2013.

The license follows.
"""
# Copyright (c) 2013, Elettra - Sincrotrone Trieste S.C.p.A.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
try:
    import pywt
    __has_pywt__ = True
except ImportError:
    __has_pywt__ = False


def munchetal_filter(im, wlevel, sigma, wname='db15'):
    """Process a sinogram image with the Munch et al. de-striping algorithm.

    Parameters
    ----------
    im : array_like
        Image data as numpy array.
    wname : {'haar', 'db1'-'db20', 'sym2'-'sym20', 'coif1'-'coif5', 'dmey'}
        The wavelet transform to use.
    wlevel : int
        Levels of the wavelet decomposition.
    sigma : float
        Cutoff frequency of the Butterworth low-pass filtering.

    Example (using tiffile.py)
    --------------------------
    >>> im = imread('original.tif')
    >>> im = munchetal_filter(im, 'db15', 4, 1.0)
    >>> imsave('filtered.tif', im)

    References
    ----------
    B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
    combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
    """
    # Wavelet decomposition:
    coeffs = pywt.wavedec2(im.astype(np.float32), wname, level=wlevel, mode="periodization")
    coeffsFlt = [coeffs[0]]
    # FFT transform of horizontal frequency bands:
    for i in range(1, wlevel + 1):
        # FFT:
        fcV = np.fft.fftshift(np.fft.fft(coeffs[i][1], axis=0))
        my, mx = fcV.shape
        # Damping of vertical stripes:
        damp = 1 - np.exp(-(np.arange(-np.floor(my / 2.), -np.floor(my / 2.) + my) ** 2) / (2 * (sigma ** 2)))
        dampprime = np.kron(np.ones((1, mx)), damp.reshape((damp.shape[0], 1))) # np.tile(damp[:, np.newaxis], (1, mx))
        fcV = fcV * dampprime
        # Inverse FFT:
        fcVflt = np.real(np.fft.ifft(np.fft.ifftshift(fcV), axis=0))
        cVHDtup = (coeffs[i][0], fcVflt, coeffs[i][2])
        coeffsFlt.append(cVHDtup)

    # Get wavelet reconstruction:
    im_f = np.real(pywt.waverec2(coeffsFlt, wname, mode="periodization"))
    # Return image according to input type:
    if (im.dtype == 'uint16'):
        # Check extrema for uint16 images:
        im_f[im_f < np.iinfo(np.uint16).min] = np.iinfo(np.uint16).min
        im_f[im_f > np.iinfo(np.uint16).max] = np.iinfo(np.uint16).max
        # Return filtered image (an additional row and/or column might be present):
        return im_f[0:im.shape[0], 0:im.shape[1]].astype(np.uint16)
    else:
        return im_f[0:im.shape[0], 0:im.shape[1]]


if not(__has_pywt__):
    munchetal_filter = None



from math import floor
import numpy as np

class VerticalShift:
    def __init__(self, radios_shape, shifts):
        """
        This class is used when a vertical translation (along the tomography
        rotation axis) occurred.

        These translations are meant "per projection" and can be due either to
        mechanical errors, or can be applied purposefully with known motor movements
        to smear rings artefacts.

        The object is initialised with an array of shifts: one shift for each projection.
        A positive shifts means that the axis has moved in the positive Z direction.
        The interpolation is done taking for a pixel (y,x) the pixel found at (y+shft,x)
        in the recorded images.

        The method apply_vertical_shifts performs the correctionson the radios.

        Parameters
        ----------
        radios_shape: tuple
            Shape of the radios chunk, in the form (n_radios, n_y, n_x)
        shifts: sequence of floats
            one shift for each projection

        Notes
        ------
        During the acquisition, there might be other translations, each of them
        orthogonal to the rotation axis.
          - A "horizontal" translation in the detector plane: this is handled
            directly in the Backprojection operation.
          - A translation along the beam direction: this one is of no concern
            for parallel-beam geometry
        """
        self.radios_shape = radios_shape
        self.shifts = shifts
        self._init_interp_coefficients()

    def _init_interp_coefficients(self):
        self.interp_infos = []
        for s in self.shifts:
            s0 = int(floor(s))
            f = s - s0
            self.interp_infos.append([s0, f])


    def _check(self, radios, iangles):
        assert np.min(iangles) >= 0
        assert np.max(iangles) < len(self.interp_infos)
        assert len(iangles) == radios.shape[0]


    def apply_vertical_shifts(self, radios, iangles, output=None):
        """
        Parameters
        ----------
        radios: a sequence of np.array
            The input radios. If the optional parameter is not given, they are modified in-place
        iangles:  a sequence of integers
            Must have the same lenght as radios.
            It contains the index at which the shift is found in `self.shifts`
            given by `shifts` argument in the initialisation of the object.
        output: a sequence of np.array, optional
            If given, it will be modified to contain the shifted radios.
            Must be of the same shape of `radios`.
        """
        self._check(radios, iangles)

        newradio = np.zeros_like(radios[0])
        for radio, ia in zip(radios, iangles):
            newradio[:] = 0
            S0, f = self.interp_infos[ia]
            s0 = S0

            if s0 > 0:
                newradio[:-s0] = radio[s0:] * (1 - f)
            elif s0 == 0:
                newradio[:] = radio[s0:] * (1 - f)
            else:
                newradio[-s0:] = radio[:s0] * (1 - f)

            s0 = S0 + 1

            if s0 > 0:
                newradio[:-s0] += radio[s0:] * f
            elif s0 == 0:
                newradio[:] += radio[s0:] * f
            else:
                newradio[-s0:] += radio[:s0] * f

            if output is None:
                radios[ia] = newradio
            else:
                output[ia] = newradio



from ..pipeline.fullfield.grouped import GroupedPipeline as FullRadiosPipeline, SinoStackPipeline
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.resources.app.fullfield_mixed was moved to nabu.pipeline.fullfield.grouped.",
    func_name="fullfield_mixed_reconstruction"
)


from bisect import bisect_left
import numpy as np
from ..io.reader import load_images_from_dataurl_dict
from ..utils import check_supported, deprecated

class FlatFieldArrays:
    """
    A class for flat-field normalization
    """

    _supported_interpolations = ["linear", "nearest"]

    def __init__(
        self,
        radios_shape: tuple,
        flats,
        darks,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
    ):
        """
        Initialize a flat-field normalization process.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where each key is the flat index, and the value is a
            numpy.ndarray of the flat image.
        darks: dict
            Dictionary where each key is the dark index, and the value is a
            numpy.ndarray of the dark image.
        radios_indices: array of int, optional
            Array containing the radios indices in the scan. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to correct flat distortions based on each radio.


        Important
        ----------
        `flats` and `darks` are expected to be a dictionary with integer keys (the flats/darks indices)
        and numpy array values.
        You can use the following helper functions: `nabu.io.reader.load_images_from_dataurl_dict`
        and `nabu.io.utils.create_dict_of_indices`


        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.
        """
        self._set_parameters(radios_shape, radios_indices, interpolation)
        self._set_flats_and_darks(flats, darks)
        self.distortion_correction = distortion_correction


    def _set_flats_and_darks(self, flats, darks):
        self._check_frames(flats, "flats", 1, 9999)
        self.n_flats = len(flats)
        self.flats = flats
        self._sorted_flat_indices = sorted(self.flats.keys())
        self._flat2arrayidx = {k: v for k, v in zip(self._sorted_flat_indices, np.arange(self.n_flats))}

        self._check_frames(darks, "darks", 1, 1)
        self.darks = darks
        self.n_darks = len(darks)
        self._sorted_dark_indices = sorted(self.darks.keys())
        self._dark = None


    def _set_radios_shape(self, radios_shape):
        if len(radios_shape) == 2:
            self.radios_shape = (1,) + radios_shape
        elif len(radios_shape) == 3:
            self.radios_shape = radios_shape
        else:
            raise ValueError("Expected radios to have 2 or 3 dimensions")
        n_radios, n_z, n_x = self.radios_shape
        self.n_radios = n_radios
        self.n_angles = n_radios
        self.shape = (n_z, n_x)


    @staticmethod
    def _check_frames(frames, frames_type, min_frames_required, max_frames_supported):
        n_frames = len(frames)
        if n_frames < min_frames_required:
            raise ValueError("Need at least %d %s" % (min_frames_required, frames_type))
        if n_frames > max_frames_supported:
            raise ValueError(
                "Flat-fielding with more than %d %s is not supported"
                % (max_frames_supported, frames_type)
            )


    def _set_parameters(self, radios_shape, radios_indices, interpolation):
        self._set_radios_shape(radios_shape)
        if radios_indices is None:
            radios_indices = np.arange(0, self.n_radios, dtype=np.int32)
        else:
            radios_indices = np.array(radios_indices, dtype=np.int32)
            if radios_indices.size != self.n_radios:
                raise ValueError(
                    "Expected radios_indices to have length %s = n_radios, but got length %d"
                    % (self.n_radios, radios_indices.size)
                )
        self.radios_indices = radios_indices
        self.interpolation = interpolation
        check_supported(
            interpolation, self._supported_interpolations, "Interpolation mode"
        )

    @staticmethod
    def get_previous_next_indices(arr, idx):
        pos = bisect_left(arr, idx)
        if pos == len(arr):  # outside range
            return (arr[-1],)
        if arr[pos] == idx:
            return (idx,)
        if pos == 0:
            return (arr[0],)
        return arr[pos - 1], arr[pos]

    @staticmethod
    def get_nearest_index(arr, idx):
        pos = bisect_left(arr, idx)
        if arr[pos] == idx:
            return idx
        return arr[pos - 1] if idx - arr[pos - 1] < arr[pos] - idx else arr[pos]

    def _get_flat_linear(self, idx, dtype=np.float32):
        prev_next = self.get_previous_next_indices(self._sorted_flat_indices, idx)
        if len(prev_next) == 1:  # current index corresponds to an acquired flat
            flat_data = self.flats[prev_next[0]]
        else:  # interpolate
            prev_idx, next_idx = prev_next
            flat_data_prev = self.flats[prev_idx]
            flat_data_next = self.flats[next_idx]
            delta = next_idx - prev_idx
            w1 = 1 - (idx - prev_idx) / delta
            w2 = 1 - (next_idx - idx) / delta
            flat_data = w1 * flat_data_prev + w2 * flat_data_next
        if flat_data.dtype != dtype:
            flat_data = np.ascontiguousarray(flat_data, dtype=dtype)
        return flat_data

    def _get_flat_nearest(self, idx, dtype=np.float32):
        idx0 = self.get_nearest_index(self._sorted_flat_indices, idx)
        flat = self.flats[idx0]
        if flat.dtype != dtype:
            flat = np.ascontiguousarray(flat, dtype=dtype)
        return flat

    def get_flat(self, idx, dtype=np.float32):
        """
        Get flat file corresponding to the corresponding index.
        If no flat file was acquired at this index, an interpolated flat
        is returned.
        """
        flat = None
        if self.interpolation == "linear":
            flat = self._get_flat_linear(idx, dtype=dtype)
        elif self.interpolation == "nearest":
            flat = self._get_flat_nearest(idx, dtype=dtype)
        else:
            raise ValueError("Unsupported interpolation method %s" % self.interpolation)
        return flat

    def get_dark(self):
        if self._dark is None:
            first_dark_idx = self._sorted_dark_indices[0]
            dark = np.ascontiguousarray(self.darks[first_dark_idx], dtype=np.float32)
            self._dark = dark
        return self._dark

    def normalize_radios(self, radios):
        """
        Apply a flat-field normalization, with the current parameters, to a stack
        of radios.
        The processing is done in-place, meaning that the radios content is overwritten.

        Parameters
        -----------
        radios: numpy.ndarray
            Radios chunk
        """
        do_flats_distortion_correction = (self.distortion_correction is not None)
        dark = self.get_dark()
        for i, idx in enumerate(self.radios_indices):
            radio_data = radios[i]
            radio_data -= dark
            flat = self.get_flat(idx)
            flat = flat - dark
            if do_flats_distortion_correction:
                flat = self.distortion_correction.estimate_and_correct(flat, radio_data)
            radios[i] = radio_data / flat
        return radios


    def normalize_single_radio(self, radio, radio_idx, dtype=np.float32):
        """
        Apply a flat-field normalization to a single projection image.
        """
        dark = self.get_dark()
        radio -= dark
        flat = self.get_flat(radio_idx, dtype=dtype)
        flat = flat - dark
        if self.distortion_correction is not None:
            flat = self.distortion_correction.estimate_and_correct(flat, radio)
        radio /= flat
        return radio


FlatField = FlatFieldArrays

class FlatFieldDataUrls(FlatField):

    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        **chunk_reader_kwargs
    ):
        """
        Initialize a flat-field normalization process with DataUrls.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where the key is the flat index, and the value is a
            silx.io.DataUrl pointing to the flat.
        darks: dict
            Dictionary where the key is the dark index, and the value is a
            silx.io.DataUrl pointing to the dark.
        radios_indices: array, optional
            Array containing the radios indices. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to


        Other Parameters
        ----------------
        The other named parameters are passed to ChunkReader(). Please read its
        documentation for more information.

        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.
        """

        flats_arrays_dict = load_images_from_dataurl_dict(flats, **chunk_reader_kwargs)
        darks_arrays_dict = load_images_from_dataurl_dict(darks, **chunk_reader_kwargs)
        super().__init__(
            radios_shape,
            flats_arrays_dict,
            darks_arrays_dict,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
        )




from math import pi
import pycuda.gpuarray as garray
import numpy as np
from silx.image.tomography import compute_fourier_filter, get_next_power
from ..cuda.kernel import CudaKernel
from ..cuda.processing import CudaProcessing
from ..utils import get_cuda_srcfile, check_supported, updiv

class SinoFilter:

    available_padding_modes = ["zeros", "edges"]

    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        extra_options=None,
        cuda_options=None,
    ):
        """
        Build a sinogram filter process.
        """
        self.cuda = CudaProcessing(**(cuda_options or {}))
        self._init_extra_options(extra_options)
        self._calculate_shapes(sino_shape)
        self._init_fft()
        self._allocate_memory(sino_shape)
        self._compute_filter(filter_name)
        self._set_padding_mode(padding_mode)
        self._init_kernels()


    def _init_extra_options(self, extra_options):
        self.extra_options = {
            "cutoff": 1.,
        }
        if extra_options is not None:
            self.extra_options.update(extra_options)

    def _set_padding_mode(self, padding_mode):
        # Compat.
        if padding_mode == "edge":
            padding_mode = "edges"
        #
        check_supported(padding_mode, self.available_padding_modes, "padding mode")
        self.padding_mode = padding_mode


    def _calculate_shapes(self, sino_shape):
        self.ndim = len(sino_shape)
        if self.ndim == 2:
            n_angles, dwidth = sino_shape
            n_sinos = 1
        elif self.ndim == 3:
            n_sinos, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Invalid sinogram number of dimensions")
        self.sino_shape = sino_shape
        self.n_angles = n_angles
        self.dwidth = dwidth
        # int() is crucial here ! Otherwise some pycuda arguments (ex. memcpy2D)
        # will not work with numpy.int64 (as for 2018.X)
        self.dwidth_padded = int(get_next_power(2*self.dwidth))
        self.sino_padded_shape = (n_angles, self.dwidth_padded)
        if self.ndim == 3:
            self.sino_padded_shape = (n_sinos, ) + self.sino_padded_shape
        sino_f_shape = list(self.sino_padded_shape)
        sino_f_shape[-1] = sino_f_shape[-1]//2+1
        self.sino_f_shape = tuple(sino_f_shape)
        #
        self.pad_left = (self.dwidth_padded - self.dwidth)//2
        self.pad_right = self.dwidth_padded - self.dwidth - self.pad_left


    def _init_fft(self):
        # Import has to be done here, otherwise scikit-cuda creates a cuda/cublas context at import
        from silx.math.fft.cufft import CUFFT
        #
        self.fft = CUFFT(
            self.sino_padded_shape,
            dtype=np.float32,
            axes=(-1,),
        )


    def _allocate_memory(self, sino_shape):
        self.d_filter_f = garray.zeros((self.sino_f_shape[-1],), np.complex64)
        self.d_sino_padded = self.fft.data_in
        self.d_sino_f = self.fft.data_out


    def set_filter(self, h_filt, normalize=True):
        """
        Set a filter for sinogram filtering.

        :param h_filt: Filter. Each line of the sinogram will be filtered with
            this filter. It has to be the Real-to-Complex Fourier Transform
            of some real filter, padded to 2*sinogram_width.
        :param normalize: Whether to normalize the filter with pi/num_angles.
        """
        if h_filt.size != self.sino_f_shape[-1]:
            raise ValueError(
                """
                Invalid filter size: expected %d, got %d.
                Please check that the filter is the Fourier R2C transform of
                some real 1D filter.
                """
                % (self.sino_f_shape[-1], h_filt.size)
            )
        if not(np.iscomplexobj(h_filt)):
            print("Warning: expected a complex Fourier filter")
        self.filter_f = h_filt
        if normalize:
            self.filter_f *= pi/self.n_angles
        self.filter_f = self.filter_f.astype(np.complex64)
        self.d_filter_f[:] = self.filter_f[:]


    def _compute_filter(self, filter_name):
        self.filter_name = filter_name or "ram-lak"
        filter_f = compute_fourier_filter(
            self.dwidth_padded,
            self.filter_name,
            cutoff=self.extra_options["cutoff"],
        )[:self.dwidth_padded//2+1]  # R2C
        self.set_filter(filter_f, normalize=True)


    def _init_kernels(self):
        fname = get_cuda_srcfile("ElementOp.cu")
        if self.ndim == 2:
            kernel_name = "inplace_complex_mul_2Dby1D"
            kernel_sig = "PPii"
        else:
            kernel_name = "inplace_complex_mul_3Dby1D"
            kernel_sig = "PPiii"
        self.mult_kernel = CudaKernel(
            kernel_name,
            filename=fname,
            signature=kernel_sig
        )
        self.kern_args = (self.d_sino_f, self.d_filter_f)
        self.kern_args += (self.d_sino_f.shape[::-1])
        self._pad_edges_kernel = CudaKernel(
            "padding_edge",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiii"
        )
        self._pad_block = (32, 32, 1)
        self._pad_grid = tuple([updiv(n, p) for n, p in zip(self.sino_padded_shape[::-1], self._pad_block)])


    def _check_array(self, arr):
        if arr.dtype != np.float32:
            raise ValueError("Expected data type = numpy.float32")
        if arr.shape != self.sino_shape:
            raise ValueError("Expected sinogram shape %s, got %s" % (self.sino_shape, arr.shape))


    def _pad_sino(self, sino):
        if self.padding_mode == "edges":
            self.d_sino_padded[:, :self.dwidth] = sino[:]
            self._pad_edges_kernel(
                self.d_sino_padded,
                self.dwidth,
                self.n_angles,
                self.dwidth_padded,
                self.n_angles,
                self.pad_left,
                self.pad_right,
                0,
                0,
                grid=self._pad_grid,
                block=self._pad_block
            )
        else: # zeros
            self.d_sino_padded.fill(0)
            if self.ndim == 2:
                self.d_sino_padded[:, :self.dwidth] = sino[:]
            else:
                self.d_sino_padded[:, :, :self.dwidth] = sino[:]


    def filter_sino(self, sino, output=None, no_output=False):
        """
        Perform the sinogram siltering.

        Parameters
        ----------
        sino: numpy.ndarray or pycuda.gpuarray.GPUArray
            Input sinogram (2D or 3D)
        output: numpy.ndarray or pycuda.gpuarray.GPUArray, optional
            Output array.
        no_output: bool, optional
            If set to True, no copy is be done. The resulting data lies
            in self.d_sino_padded.
        """
        self._check_array(sino)
        # copy2d/copy3d
        self._pad_sino(sino)
        # FFT
        self.fft.fft(self.d_sino_padded, output=self.d_sino_f)

        # multiply padded sinogram with filter in the Fourier domain
        self.mult_kernel(*self.kern_args) # TODO tune block size ?

        # iFFT
        self.fft.ifft(self.d_sino_f, output=self.d_sino_padded)

        # return
        if no_output:
            return self.d_sino_padded
        if output is None:
            res = np.zeros(self.sino_shape, dtype=np.float32)
            # can't do memcpy2d D->H ? (self.d_sino_padded[:, w]) I have to get()
            sino_ref = self.d_sino_padded.get()
        else:
            res = output
            sino_ref = self.d_sino_padded
        if self.ndim == 2:
            res[:] = sino_ref[:, :self.dwidth]
        else:
            res[:] = sino_ref[:, :, :self.dwidth]
        return res

    __call__ = filter_sino



import os
from typing import Union
import numpy as np
from silx.io.dictdump import h5todict
from tomoscan.io import HDF5File
from .utils import get_compacted_dataslices, convert_dict_values
from ..misc.binning import get_binning_function
from ..utils import subsample_dict, get_3D_subregion
try:
    from silx.third_party.EdfFile import EdfFile
except ImportError:
    EdfFile = None

class Reader:
    """
    Abstract class for various file readers.
    """

    def __init__(self, sub_region=None):
        """
        Parameters
        ----------
        sub_region: tuple, optional
            Coordinates in the form (start_x, end_x, start_y, end_y), to read
            a subset of each frame. It can be used for Regions of Interest (ROI).
            Indices start at zero !
        """
        self._set_default_parameters(sub_region)

    def _set_default_parameters(self, sub_region):
        self._set_subregion(sub_region)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_x, end_x, start_y, end_y = sub_region
            self.start_x = start_x
            self.end_x = end_x
            self.start_y = start_y
            self.end_y = end_y
        else:
            self.start_x = 0
            self.end_x = None
            self.start_y = 0
            self.end_y = None


    def get_data(self, data_url):
        """
        Get data from a silx.io.url.DataUrl
        """
        raise ValueError("Base class")


    def release(self):
        """
        Release the file if needed.
        """
        pass


class NPReader(Reader):

    multi_load = True

    def __init__(self, sub_region=None, mmap=True):
        """
        Reader for NPY/NPZ files. Mostly used for internal development.
        Please refer to the documentation of nabu.io.reader.Reader
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}
        self._set_mmap(mmap)

    def _set_mmap(self, mmap):
        self.mmap_mode = "r" if mmap else None

    def _open(self, data_url):
        file_path = data_url.file_path()
        file_ext = self._get_file_type(file_path)
        if file_ext == "npz":
            if file_path not in self._file_desc:
                self._file_desc[file_path] = np.load(file_path, mmap_mode=self.mmap_mode)
            data_ref = self._file_desc[file_path][data_url.data_path()]
        else:
            data_ref = np.load(file_path, mmap_mode=self.mmap_mode)
        return data_ref

    @staticmethod
    def _get_file_type(fname):
        if fname.endswith(".npy"):
            return "npy"
        elif fname.endswith(".npz"):
            return "npz"
        else:
            raise ValueError("Not a numpy file: %s" % fname)

    def get_data(self, data_url):
        data_ref = self._open(data_url)
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = data_ref[self.start_y:self.end_y, self.start_x:self.end_x]
        else:
            res = data_ref[data_slice, self.start_y:self.end_y, self.start_x:self.end_x]
        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                fdesc.close()
                self._file_desc[fname] = None

    def __del__(self):
        self.release()


class EDFReader(Reader):

    multi_load = False # not implemented

    def __init__(self, sub_region=None):
        """
        A class for reading series of EDF Files.
        Multi-frames EDF are not supported.
        """
        if EdfFile is None:
            raise ImportError("Need EdfFile to use this reader")
        super().__init__(sub_region=sub_region)

    def read(self, fname):
        E = EdfFile(fname, "r")
        if self.sub_region is not None:
            data = E.GetData(
                0,
                Pos=(self.start_x, self.start_y),
                Size=(self.end_x - self.start_x, self.end_y - self.start_y),
            )
        else:
            data = E.GetData(0)
        E.File.close()
        return data

    def get_data(self, data_url):
        return self.read(data_url.file_path())


class HDF5Reader(Reader):

    multi_load = True

    def __init__(self, sub_region=None):
        """
        A class for reading a HDF5 File.
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}


    def _open(self, file_path):
        if file_path not in self._file_desc:
            self._file_desc[file_path] = HDF5File(file_path, "r", swmr=True)


    def get_data(self, data_url):
        file_path = data_url.file_path()
        self._open(file_path)
        h5dataset = self._file_desc[file_path][data_url.data_path()]
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = h5dataset[self.start_y:self.end_y, self.start_x:self.end_x]
        else:
            res = h5dataset[data_slice, self.start_y:self.end_y, self.start_x:self.end_x]
        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                try:
                    fdesc.close()
                    self._file_desc[fname] = None
                except Exception as exc:
                    print("Error while closing %s: %s" % (fname, str(exc)))

    def __del__(self):
        self.release()


class HDF5Loader:
    """
    An alternative class to HDF5Reader where information is first passed at class instantiation
    """
    def __init__(self, fname, data_path, sub_region=None, pre_allocate=True, dtype="f"):
        self.fname = fname
        self.data_path = data_path
        self._set_subregion(sub_region)
        self.data = None
        self._loaded = False
        if pre_allocate:
            expected_shape = get_hdf5_dataset_shape(fname, data_path, sub_region=sub_region)
            self.data = np.zeros(expected_shape, dtype=dtype)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_z, end_z, start_y, end_y, start_x, end_x = sub_region
            self.start_x, self.end_x = start_x, end_x
            self.start_y, self.end_y = start_y, end_y
            self.start_z, self.end_z = start_z, end_z
        else:
            self.start_x, self.end_x = None, None
            self.start_y, self.end_y = None, None
            self.start_z, self.end_z = None, None


    def load_data(self, force_load=False):
        if self._loaded and not force_load:
            return self.data
        with HDF5File(self.fname, "r") as fdesc:
            if self.data is None:
                self.data = fdesc[self.data_path][self.start_z:self.end_z, self.start_y:self.end_y, self.start_x:self.end_x]
            else:
                self.data[:] = fdesc[self.data_path][self.start_z:self.end_z, self.start_y:self.end_y, self.start_x:self.end_x]
        self._loaded = True
        return self.data


class ChunkReader:
    """
    A reader of chunk of images.
    """

    def __init__(
        self,
        files,
        sub_region: Union[tuple, None] = None,
        pre_allocate=True,
        data_buffer=None,
        convert_float=False,
        shape=None,
        dtype=None,
        binning=None,
        dataset_subsampling=None
    ):
        """
        Initialize a "ChunkReader". A chunk is a stack of images.

        Parameters
        ----------
        files: dict
            Dictionary where the key is the file/data index, and the value is a
            silx.io.url.DataUrl pointing to the data.
            The dict must contain only the files which shall be used !
            Note: the shape and data type is infered from the first data file.
        sub_region: tuple, optional
            If provided, this must be a tuple in the form
            (start_x, end_x, start_y, end_y). Each image will be cropped to this
            region. This is used to specify a chunk of files.
            Each of the parameters can be None, in this case the default start
            and end are taken in each dimension.
        pre_allocate: bool
            Whether to pre-allocate data before reading.
        data_buffer: array-like, optional
            If `pre_allocate` is set to False, this parameter has to be provided.
            It is an array-like object which will hold the data.
        convert_float: bool
            Whether to convert data to float32, regardless of the input data type.
        shape: tuple, optional
            Shape of each image. If not provided, it is inferred from the first image
            in the collection.
        dtype: `numpy.dtype`, optional
            Data type of each image. If not provided, it is inferred from the first image
            in the collection.
        binning: int or tuple of int, optional
            Whether to bin the data. If multi-dimensional binning is done,
            the parameter must be in the form (binning_x, binning_y).
            Each image will be binned by these factors.
        dataset_subsampling: int, optional
            Whether to subsample the dataset. If an integer `n` is provided,
            then one image out of `n` will be read.

        Notes
        ------
        The files are provided as a collection of `silx.io.DataURL`. The file type
        is inferred from the extension.

        Binning is different from subsampling. Using binning will not speed up
        the data retrieval (quite the opposite), since the whole (subregion of) data
        is read and then binning is performed.
        """
        self._get_reader_class(files)
        self.dataset_subsampling = dataset_subsampling
        self._set_files(files)
        self._get_shape_and_dtype(shape, dtype, binning)
        self._set_subregion(sub_region)
        self._init_reader()
        self._loaded = False
        self.convert_float = convert_float
        if convert_float:
            self.out_dtype = np.float32
        else:
            self.out_dtype = self.dtype
        if not((data_buffer is not None) ^ (pre_allocate is True)):
            raise ValueError("Please provide either 'data_buffer' or 'pre_allocate'")
        self.files_data = data_buffer
        if data_buffer is not None:
            # overwrite out_dtype
            self.out_dtype = data_buffer.dtype
            if data_buffer.shape != self.shape:
                raise ValueError("Expected shape %s but got %s" % (self.shape, data_buffer.shape))
        if pre_allocate:
            self.files_data = np.zeros(self.chunk_shape, dtype=self.out_dtype)
        if (self.binning is not None) and (np.dtype(self.out_dtype).kind in ["u", "i"]):
            raise ValueError("Output datatype cannot be integer when using binning. Please set the 'convert_float' parameter to True or specify a 'data_buffer'.")


    def _set_files(self, files):
        if len(files) == 0:
            raise ValueError("Expected at least one data file")
        self.n_files = len(files)
        self.files = files
        self._sorted_files_indices = sorted(files.keys())
        self._fileindex_to_idx = dict.fromkeys(self._sorted_files_indices)
        self._configure_subsampling()


    def _infer_file_type(self, files):
        fname = files[sorted(files.keys())[0]].file_path()
        ext = os.path.splitext(fname)[-1].replace(".", "")
        if ext not in Readers:
            raise ValueError(
                "Unknown file format %s. Supported formats are: %s"
                % (ext, str(Readers.keys()))
            )
        return ext

    def _get_reader_class(self, files):
        ext = self._infer_file_type(files)
        reader_class = Readers[ext]
        self._reader_class = reader_class


    def _get_shape_and_dtype(self, shape, dtype, binning):
        if shape is None or dtype is None:
            shape, dtype = self._infer_shape_and_dtype()
        assert len(shape) == 2, "Expected the shape of an image (2-tuple)"
        self.shape_total = shape
        self.dtype = dtype
        self._set_binning(binning)


    def _configure_subsampling(self):
        dataset_subsampling = self.dataset_subsampling
        self.files_subsampled = self.files
        if dataset_subsampling is not None and dataset_subsampling > 1:
            self.n_files //= dataset_subsampling
            self.files_subsampled = subsample_dict(self.files, dataset_subsampling)
            if not(self._reader_class.multi_load):
                # 3D loading not supported for this reader.
                # Data is loaded frames by frame, so subsample directly self.files
                self.files = self.files_subsampled
                self._sorted_files_indices = sorted(self.files.keys())
                self._fileindex_to_idx = dict.fromkeys(self._sorted_files_indices)


    def _infer_shape_and_dtype(self):
        self._reader_entire_image = self._reader_class()
        first_file_dataurl = self.files[self._sorted_files_indices[0]]
        first_file_data = self._reader_entire_image.get_data(first_file_dataurl)
        return first_file_data.shape, first_file_data.dtype

    def _set_subregion(self, sub_region):
        sub_region = sub_region or (None, None, None, None)
        start_x, end_x, start_y, end_y = sub_region
        if start_x is None:
            start_x = 0
        if start_y is None:
            start_y = 0
        if end_x is None:
            end_x = self.shape_total[1]
        if end_y is None:
            end_y = self.shape_total[0]
        self.sub_region = (start_x, end_x, start_y, end_y)
        self.shape = (end_y - start_y, end_x - start_x)
        if self.binning is not None:
            self.shape = (self.shape[0] // self.binning[1], self.shape[1] // self.binning[0])
        self.chunk_shape = (self.n_files,) + self.shape

    def _init_reader(self):
        # instantiate reader with user params
        self.file_reader = self._reader_class(
            sub_region=self.sub_region
        )


    def _set_binning(self, binning):
        if binning is None:
            self.binning = None
            return
        if np.isscalar(binning):
            binning = (binning, binning)
        else:
            assert len(binning) == 2, "Expected binning in the form (binning_x, binning_y)"
        if binning[0] == 1 and binning[1] == 1:
            self.binning = None
            return
        for b in binning:
            if int(b) != b:
                raise ValueError("Expected an integer number for binning values, but got %s" % binning)
            # Current limitation
            if b not in [1, 2, 3]:
                raise NotImplementedError("Currently, binning can only be 2 or 3 in each dimension")
            #
        self.binning = binning
        binning_function = get_binning_function(binning[::-1]) # user provides (binning_x, binning_y)
        if binning_function is None:
            raise NotImplementedError("Binning factor %s is not implemented yet" % str(binning))
        self.binning_function = binning_function


    def get_data(self, file_url):
        """
        Get the data associated to a file url.
        """
        arr = self.file_reader.get_data(file_url)
        if self.binning is not None:
            if arr.ndim == 2:
                arr = self.binning_function(arr)
            else:
                nz = arr.shape[0]
                res = np.zeros((nz, ) + self.binning_function(arr[0]).shape, dtype="f")
                for i in range(nz):
                    res[i] = self.binning_function(arr[i])
                arr = res
        return arr


    def _load_single(self):
        for i, fileidx in enumerate(self._sorted_files_indices):
            file_url = self.files[fileidx]
            self.files_data[i] = self.get_data(file_url)
            self._fileindex_to_idx[fileidx] = i


    def _load_multi(self):
        urls_compacted = get_compacted_dataslices(self.files, subsampling=self.dataset_subsampling)
        loaded = {}
        start_idx = 0
        for idx in self._sorted_files_indices:
            url = urls_compacted[idx]
            url_str = str(url)
            is_loaded = loaded.get(url_str, False)
            if is_loaded:
                continue
            ds = url.data_slice()
            delta_z = ds.stop - ds.start
            if ds.step is not None and ds.step > 1:
                delta_z //= ds.step
            end_idx = start_idx + delta_z
            self.files_data[start_idx:end_idx] = self.get_data(url)
            start_idx += delta_z
            loaded[url_str] = True


    def load_files(self, overwrite: bool=False):
        """
        Load the files whose links was provided at class instantiation.

        Parameters
        -----------
        overwrite: bool, optional
            Whether to force reloading the files if already loaded.
        """
        if self._loaded and not(overwrite):
            raise ValueError("Radios were already loaded. Call load_files(overwrite=True) to force reloading")
        if self.file_reader.multi_load:
            self._load_multi()
        else:
            self._load_single()
        self._loaded = True

    load_data = load_files

    @property
    def data(self):
        return self.files_data


Readers = {
    "edf": EDFReader,
    "hdf5": HDF5Reader,
    "h5": HDF5Reader,
    "nx": HDF5Reader,
    "npz": NPReader,
    "npy": NPReader,
}


def load_images_from_dataurl_dict(data_url_dict, **chunk_reader_kwargs):
    """
    Load a dictionary of dataurl into numpy arrays.

    Parameters
    ----------
    data_url_dict: dict
        A dictionary where the keys are integers (the index of each image in the dataset),
        and the values are numpy.ndarray (data_url_dict).

    Other parameters
    -----------------
    chunk_reader_kwargs: params
        Named parameters passed to `nabu.io.reader.ChunkReader`.

    Returns
    --------
    res: dict
        A dictionary where the keys are the same as `data_url_dict`, and the values are numpy arrays.
    """
    chunk_reader = ChunkReader(data_url_dict, **chunk_reader_kwargs)
    img_dict = {}
    for img_idx, img_url in data_url_dict.items():
        img_dict[img_idx] = chunk_reader.get_data(img_url)
    return img_dict


def load_images_stack_from_hdf5(fname, h5_data_path, sub_region=None):
    """
    Load a 3D dataset from a HDF5 file.

    Parameters
    -----------
    fname: str
        File path
    h5_data_path: str
        Data path within the HDF5 file
    sub_region: tuple, optional
        Tuple indicating which sub-volume to load, in the form (xmin, xmax, ymin, ymax, zmin, zmax)
        where the 3D dataset has the python shape (N_Z, N_Y, N_X).
        This means that the data will be loaded as `data[zmin:zmax, ymin:ymax, xmin:xmax]`.
    """
    xmin, xmax, ymin, ymax, zmin, zmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        data = d_ptr[zmin:zmax, ymin:ymax, xmin:xmax]
    return data


def get_hdf5_dataset_shape(fname, h5_data_path, sub_region=None):
    zmin, zmax, ymin, ymax, xmin, xmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        shape = d_ptr.shape
    n_z, n_y, n_x = shape
    # perhaps there is more elegant
    res_shape = []
    for n, bounds in zip([n_z, n_y, n_x], ((zmin, zmax), (ymin, ymax), (xmin, xmax))):
        res_shape.append(np.arange(n)[bounds[0]:bounds[1]].size)
    return tuple(res_shape)


def check_virtual_sources_exist(fname, data_path):
    with HDF5File(fname, "r") as f:
        if data_path not in f:
            print("No dataset %s in file %s" % (data_path, fname))
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(os.path.dirname(dptr.file.filename), vsource.file_name)
            if not os.path.isfile(vsource_fname):
                print("No such file: %s" % vsource_fname)
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                print("Error with virtual source %s" % vsource_fname)
                return False
    return True


def import_h5_to_dict(h5file, h5path, asarray=False):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing "None" with None

    Parameters
    -----------
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    asarray: bool, optional
        Whether to convert each numeric value to an 0D array. Default is False.
    """
    dic = h5todict(h5file, path=h5path, asarray=asarray)
    modified_dic = convert_dict_values(
        dic,
        {"None": None},
        bytes_tostring=True
    )
    return modified_dic



import os.path as path
import numpy as np
import pytest
from nabu.testutils import get_data
from nabu.cuda.utils import __has_pycuda__
from nabu.preproc.sinogram import SinoNormalization
if __has_pycuda__:
    from nabu.preproc.sinogram_cuda import CudaSinoNormalization
    import pycuda.gpuarray as garray


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("sino_refill.npy")
    cls.tol = 1e-7


@pytest.mark.usefixtures("bootstrap")
class TestSinoNormalization:

    def test_sino_normalization(self):
        sino_proc = SinoNormalization(
            kind="chebyshev", sinos_shape=self.sino.shape
        )
        sino = self.sino.copy()
        sino_proc.normalize(sino)


    @pytest.mark.skipif(not(__has_pycuda__), reason="Need pycuda for sinogram normalization with cuda backend")
    def test_dff_cuda(self):
        sino_proc = SinoNormalization(
            kind="chebyshev", sinos_shape=self.sino.shape
        )
        sino = self.sino.copy()
        ref = sino_proc.normalize(sino)

        cuda_sino_proc = CudaSinoNormalization(
            kind="chebyshev", sinos_shape=self.sino.shape
        )
        d_sino = garray.to_gpu(self.sino)
        cuda_sino_proc.normalize(d_sino)
        res = d_sino.get()

        assert np.max(np.abs(res - ref)) < self.tol




import numpy as np
from ..utils import get_cuda_srcfile, updiv, check_supported
from .kernel import CudaKernel
from .processing import CudaProcessing
import pycuda.gpuarray as garray


class CudaPadding:
    """
    A class for performing padding on GPU
    """
    supported_modes = ["constant", "edge", "reflect", "symmetric", "wrap"]

    def __init__(self, shape, pad_width, mode="constant", cuda_options=None, **kwargs):
        """
        Initialize a CudaPadding object.

        Parameters
        ----------
        shape: tuple
            Image shape
        pad_width: tuple
            Padding width for each axis. Please see the documentation of numpy.pad().
            It can also be a tuple of two numpy arrays for generic coordinate transform.
        mode: str
            Padding mode

        Other parameters
        ----------------
        constant_values: tuple
            Tuple containing the values to fill when mode="constant".
        """
        if len(shape) != 2:
            raise ValueError("This class only works on images")
        self.shape = shape
        self._set_mode(mode, **kwargs)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._get_padding_arrays(pad_width)
        self._init_cuda_coordinate_transform()


    def _set_mode(self, mode, **kwargs):
        if mode == "edges":
            mode = "edge"
        check_supported(mode, self.supported_modes, "padding mode")
        self.mode = mode
        self._kwargs = kwargs


    def _get_padding_arrays(self, pad_width):
        self.pad_width = pad_width
        if isinstance(pad_width, tuple) and isinstance(pad_width[0], np.ndarray):
            # user-defined coordinate transform
            if len(pad_width) != 2:
                raise ValueError(
                    "pad_width must be either a scalar, a tuple in the form ((a, b), (c, d)), or a tuple of two numpy arrays"
                )
            if self.mode == "constant":
                raise ValueError("Custom coordinate transform does not work with mode='constant'")
            self.mode = "custom"
            self.coords_rows, self.coords_cols = pad_width
        else:
            if self.mode == "constant":
                # no need for coordinate transform here
                constant_values = self._kwargs.get("constant_values", 0)
                self.padded_array_constant = np.pad(
                    np.zeros(self.shape, dtype="f"),
                    self.pad_width,
                    mode="constant",
                    constant_values=constant_values
                )
                self.padded_shape = self.padded_array_constant.shape
                return
            R, C = np.indices(self.shape, dtype=np.int32)
            self.coords_rows = np.pad(R, self.pad_width, mode=self.mode)
            self.coords_cols = np.pad(C, self.pad_width, mode=self.mode)
        self.padded_shape = self.coords_rows.shape


    def _init_cuda_coordinate_transform(self):
        if self.mode == "constant":
            self.d_padded_array_constant = garray.to_gpu(self.padded_array_constant)
            return
        self._coords_transform_kernel = CudaKernel(
            "coordinate_transform",
            filename=get_cuda_srcfile("padding.cu"),
            signature="PPPPiii",
        )
        self._coords_transform_block = (32, 32, 1)
        self._coords_transform_grid = [
            updiv(a, b)
            for a, b in zip(self.padded_shape[::-1], self._coords_transform_block)
        ]
        self.d_coords_rows = garray.to_gpu(self.coords_rows)
        self.d_coords_cols = garray.to_gpu(self.coords_cols)


    def _pad_constant(self, image, output):
        pad_y, pad_x = self.pad_width
        self.d_padded_array_constant[pad_y[0]:pad_y[0]+self.shape[0], pad_x[0]: pad_x[0] + self.shape[1]] = image[:]
        output[:] = self.d_padded_array_constant[:]
        return output


    def pad(self, image, output=None):
        """
        Pad an array.

        Parameters
        ----------
        image: pycuda.gpuarray.GPUArray
            Image to pad
        output: pycuda.gpuarray.GPUArray, optional
            Output image. If provided, must be in the expected shape.
        """
        if output is None:
            output = self.cuda_processing.allocate_array("d_output", self.padded_shape)
        if self.mode == "constant":
            return self._pad_constant(image, output)
        self._coords_transform_kernel(
            image,
            output,
            self.d_coords_cols,
            self.d_coords_rows,
            np.int32(self.shape[1]),
            np.int32(self.padded_shape[1]),
            np.int32(self.padded_shape[0]),
            grid=self._coords_transform_grid,
            block=self._coords_transform_block
        )
        return output



import os
import numpy as np
import pytest
import h5py
try:
    from algotom.prep.conversion import convert_sinogram_360_to_180
    __has_algotom__ = True
except ImportError:
    __has_algotom__ = False
from nabu.testutils import compare_arrays, utilstest
from nabu.reconstruction.sinogram import SinoBuilder
from nabu.cuda.utils import __has_pycuda__

if __has_pycuda__:
    from nabu.cuda.utils import get_cuda_context
    import pycuda.gpuarray as garray
    from nabu.reconstruction.sinogram_cuda import CudaSinoBuilder


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    sino, sino_ref, cor = get_data_h5("halftomo_new.h5")
    cls.sino = sino
    cls.radios = convert_sino_to_radios_stack(sino)
    cls.sino_ref = sino_ref
    cls.rot_center = cor
    cls.tol = 5e-3
    if __has_pycuda__:
        cls.ctx = get_cuda_context()


def convert_sino_to_radios_stack(sino):
    return np.moveaxis(np.tile(sino, (1, 1, 1)), 1, 0)


def get_data_h5(*dataset_path):
    dataset_relpath = os.path.join(*dataset_path)
    dataset_path = utilstest.getfile(dataset_relpath)
    with h5py.File(dataset_path, "r") as hf:
        sino = hf["entry/radio/results/data"][()]
        sino_extended_ref = hf["entry/sino/results/data"][()]
        cor = hf["entry/sino/configuration/configuration/rotation_axis_position"][()]
    return sino, sino_extended_ref, cor


def generate_halftomo_sinogram_algotom(sino, cor):
    """
    Generate the 180 degrees sinogram with algotom.
    """
    n_angles, dwidth = sino.shape
    # If the sinogram has an even number of projections n_a that are exactly matched,
    # then the resulting sinogram should have n_a//2 angles.
    #
    #   0                                   0       180
    #   1                                   1       181
    #   2                                   2       182
    #   ...             -- convert -->      ...     ...
    #   180                                 179     360-1
    #   181
    #   ...
    #   360-1
    #
    # Yet for some reason algotom yields n_a//2 + 1 angles, because the "180 degrees"
    # radio is used in both "sino_top" and "sino_bottom".
    # Thus we have to cheat a little bit and use an odd number of projections.
    if n_angles % 2 == 0:
        sino = np.vstack([sino, sino[-1]]) # could even be zeros for the last line
    # In nabu we use the following overlap width.
    # algotom default is 2 * (dwidth - cor - 1), which yields 2 extra pixels
    overlap_width = 2*(dwidth - int(cor))
    sino_halftomo, new_cor = convert_sinogram_360_to_180(
        sino,
        (overlap_width, 1), # 1 means that CoR is on the right
        norm=False
    )
    if n_angles % 2 == 0:
        sino_halftomo = sino_halftomo[:-1, :]
    return sino_halftomo



@pytest.mark.usefixtures("bootstrap")
class TestHalftomo:

    def _build_sinos(self, radios, output=None, backend="python"):
        sinobuilder_cls = CudaSinoBuilder if backend == "cuda" else SinoBuilder
        sino_builder = sinobuilder_cls(
            radios_shape=radios.shape, rot_center=self.rot_center, halftomo=True
        )
        sinos_halftomo = sino_builder.radios_to_sinos(radios, output=output)
        return sinos_halftomo


    def _check_result(self, sino, test_description):
        _, err = compare_arrays(
            sino, self.sino_ref, self.tol, return_residual=True
        )
        assert err < self.tol, "Something wrong with %s" % test_description


    def test_halftomo(self):
        sinos_halftomo = self._build_sinos(self.radios, backend="python")
        self._check_result(sinos_halftomo[0], "SinoBuilder.radios_to_sino, halftomo=True")


    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    def test_cuda_halftomo(self):
        d_radios = garray.to_gpu(self.radios)
        d_sinos = garray.zeros((1,) + self.sino_ref.shape, "f")
        self._build_sinos(d_radios, output=d_sinos, backend="cuda")
        self._check_result(d_sinos.get()[0], "CudaSinoBuilder.radios_to_sino, halftomo=True")


    def _get_sino_with_odd_nprojs(self):
        n_a = self.sino.shape[0]
        # dummy line inserted in the middle,
        # so that result should match reference sinogram with an even number of angles
        sino_odd = np.vstack([
            self.sino[:n_a//2],
            self.sino[-1],  # dummy line
            self.sino[n_a//2:]
        ])
        return sino_odd

    def test_halftomo_odd(self):
        sino_odd = self._get_sino_with_odd_nprojs()
        radios = convert_sino_to_radios_stack(sino_odd)
        assert radios.shape[0] & 1, "Radios must have a odd number of angles"
        sinos = self._build_sinos(radios, backend="python")
        sino_halftomo = sinos[0][:-1, :]
        self._check_result(sino_halftomo, "SinoBuilder.radios_to_sino, halftomo=True, odd number of projs")


    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    def test_cuda_halftomo_odd(self):
        sino_odd = self._get_sino_with_odd_nprojs()
        radios = convert_sino_to_radios_stack(sino_odd)
        assert radios.shape[0] & 1, "Radios must have a odd number of angles"
        d_radios = garray.to_gpu(radios)
        d_out = garray.zeros((1, self.sino_ref.shape[0]+1, self.sino_ref.shape[1]), "f")
        self._build_sinos(d_radios, output=d_out, backend="cuda")
        sino_halftomo = d_out.get()[0][:-1, :]
        self._check_result(sino_halftomo, "CudaSinoBuilder.radios_to_sino, halftomo=True, odd number of projs")


    @staticmethod
    def _flip_array(arr):
        if arr.ndim == 2:
            return np.fliplr(arr)
        res = np.zeros_like(arr)
        for i in range(arr.shape[0]):
            res[i] = np.fliplr(arr[i])
        return res

    def test_halftomo_left(self):
        na, nz, nx = self.radios.shape
        left_cor = nx - 1 - self.rot_center
        radios = self._flip_array(self.radios)
        sino_builder = SinoBuilder(
            radios_shape=radios.shape, rot_center=left_cor, halftomo=True
        )
        sinos_halftomo = sino_builder.radios_to_sinos(radios)
        _, err = compare_arrays(
            sinos_halftomo[0],
            self._flip_array(self.sino_ref),
            self.tol,
            return_residual=True,
        )
        assert (
            err < self.tol
        ), "Something wrong with SinoBuilder.radios_to_sino, halftomo=True"

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    def test_cuda_halftomo_left(self):
        na, nz, nx = self.radios.shape
        left_cor = nx - 1 - self.rot_center
        radios = self._flip_array(self.radios)
        sino_processing = CudaSinoBuilder(
            radios_shape=radios.shape, rot_center=left_cor, halftomo=True
        )
        d_radios = garray.to_gpu(radios)
        d_sinos = garray.zeros(sino_processing.sinos_halftomo_shape, "f")
        sino_processing.radios_to_sinos(d_radios, output=d_sinos)
        sino_halftomo = d_sinos.get()[0]
        _, err = compare_arrays(
            sino_halftomo, self._flip_array(self.sino_ref),
            self.tol, return_residual=True
        )
        assert (
            err < self.tol
        ), "Something wrong with SinoBuilder.radios_to_sino, halftomo=True"



import numpy as np

"""
Specialized functions for image binning.
"""


# 5 ms on 2048*2047 image (f32 or u16): 1.7 GB/s
# should be enough to be faster than reading a file from disk
# Can be accelerated well below 1ms with numba.njit(parallel=True)
def binning2(img, out_dtype=np.float32):
    """
    Perform a 2X2 binning on an image.
    """
    s = img.shape
    shp = (s[0] - (s[0] % 2), s[1] - (s[1] % 2))
    sub_img = img[:shp[0], :shp[1]]
    out_shp = (shp[0] // 2, shp[1] // 2)
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[::2, ::2].astype(out_dtype) + sub_img[1::2, ::2] + sub_img[::2, 1::2] + sub_img[1::2, 1::2]
    res *= 0.25
    return res


def binning2_horiz(img, out_dtype=np.float32):
    """
    Perform a 2-binning horizontally on an image.
    """
    s = img.shape
    shp = (s[0], s[1] - (s[1] % 2))
    sub_img = img[:, :shp[1]]
    out_shp = (shp[0], shp[1] // 2)
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[:, ::2].astype(out_dtype) + sub_img[:, 1::2]
    res *= 0.5
    return res


def binning2_vertic(img, out_dtype=np.float32):
    """
    Perform a 2-binning vertically on an image.
    """
    s = img.shape
    shp = (s[0] - (s[0] % 2), s[1])
    sub_img = img[:shp[0], :]
    out_shp = (shp[0] //2, shp[1])
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[::2, :].astype(out_dtype) + sub_img[1::2, :]
    res *= 0.5
    return res


def binning3(img, out_dtype=np.float32):
    """
    Perform a 3X3 binning on an image.
    """
    s = img.shape
    shp = (s[0] - (s[0] % 3), s[1] - (s[1] % 3))
    sub_img = img[:shp[0], :shp[1]]
    out_shp = (shp[0] // 3, shp[1] // 3)
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[::3, ::3].astype(out_dtype)  + sub_img[1::3, ::3]  + sub_img[2::3, ::3] \
           + sub_img[::3, 1::3] + sub_img[1::3, 1::3] + sub_img[2::3, 1::3] \
           + sub_img[::3, 2::3] + sub_img[1::3, 2::3] + sub_img[2::3, 2::3]
    res /= 9.
    return res



def binning3_horiz(img, out_dtype=np.float32):
    """
    Perform a 3-binning on an image horizontally
    """
    s = img.shape
    shp = (s[0], s[1] - (s[1] % 3))
    sub_img = img[:, :shp[1]]
    out_shp = (shp[0], shp[1] // 3)
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[:, ::3].astype(out_dtype)  + sub_img[:, 1::3] + sub_img[:, 2::3]
    res /= 3.
    return res



def binning3_vertic(img, out_dtype=np.float32):
    """
    Perform a 3-binning on an image vertically.
    """
    s = img.shape
    shp = (s[0] - (s[0] % 3), s[1])
    sub_img = img[:shp[0], :]
    out_shp = (shp[0] // 3, shp[1])
    res = np.zeros(out_shp, dtype=out_dtype)
    # astype() prevents overflows for integer-like input arrays, at the expanse of 50% lower performance
    res[:] = sub_img[::3, :].astype(out_dtype)  + sub_img[1::3, :] + sub_img[2::3, :]
    res /= 3.
    return res





def get_binning_function(binning):
    """
    Determine the binning function to use.
    """
    binning_functions = {
        (2, 2): binning2,
        (2, 1): binning2_vertic,
        (1, 2): binning2_horiz,
        (3, 3): binning3,
        (3, 1): binning3_vertic,
        (1, 3): binning3_horiz,
        (2, 3): None, # current limitation
        (3, 2): None, # current limitation
    }
    if binning not in binning_functions:
        raise ValueError("Could not get a function for binning factor %s" % binning)
    return binning_functions[binning]



import numpy as np
import pycuda.gpuarray as garray
from ..utils import get_cuda_srcfile
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from .rings import MunchDeringer
try:
    from pycudwt import Wavelets
    __have_pycudwt__ = True
except ImportError:
    __have_pycudwt__ = False
try:
    from skcuda.fft import Plan
    from skcuda.fft import fft as cufft
    from skcuda.fft import ifft as cuifft
    __have_skcuda__ = True
except ImportError:
    __have_skcuda__ = False


class CudaMunchDeringer(MunchDeringer):

    def __init__(self, sigma, sinos_shape=None, radios_shape=None, levels=None, wname='db15', cuda_options=None):
        """
        Initialize a "Munch Et Al" sinogram deringer with the Cuda backend.
        See References for more information.

        Parameters
        -----------
        sigma: float
            Standard deviation of the damping parameter. The higher value of sigma,
            the more important the filtering effect on the rings.
        levels: int, optional
            Number of wavelets decomposition levels.
            By default (None), the maximum number of decomposition levels is used.
        wname: str, optional
            Default is "db15" (Daubechies, 15 vanishing moments)
        sinos_shape: tuple, optional
            Shape of the sinogram (or sinograms stack).
            This class requires either sinos_shape or radios_shape.
        radios_shape: tuple, optional
            Shape of the projection (or projections stack)
            This class requires either sinos_shape or radios_shape.

        References
        ----------
        B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
        combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
        """
        super().__init__(
            sigma, levels=levels, wname=wname, sinos_shape=sinos_shape, radios_shape=radios_shape
        )
        self._check_can_use_wavelets()
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_pycudwt()
        self._init_fft()
        self._setup_fw_kernel()

    def _check_can_use_wavelets(self):
        if not(__have_pycudwt__ and __have_skcuda__):
            raise ValueError("Needs pycudwt and scikit-cuda to use this class")


    def _init_fft(self):
        self._fft_plans = {}
        for level, d_vcoeff in self._d_vertical_coeffs.items():
            n_angles, dwidth = d_vcoeff.shape
            # Batched vertical 1D FFT - need advanced data layout
            # http://docs.nvidia.com/cuda/cufft/#advanced-data-layout
            p_f = Plan(
                (n_angles,),
                np.float32, np.complex64,
                batch=dwidth,
                inembed=np.int32([0]),
                istride=dwidth,
                idist=1,
                onembed=np.int32([0]),
                ostride=dwidth,
                odist=1
            )
            p_i = Plan(
                (n_angles,),
                np.complex64, np.float32,
                batch=dwidth,
                inembed=np.int32([0]),
                istride=dwidth,
                idist=1,
                onembed=np.int32([0]),
                ostride=dwidth,
                odist=1
            )
            self._fft_plans[level] = {"forward": p_f, "inverse": p_i}

    def _init_pycudwt(self):
        if self.levels is None:
            self.levels = 100 # will be clipped by pycudwt
        self.sino_shape = self.sinos_shape[1:]
        self.cudwt = Wavelets(np.zeros(self.sino_shape, "f"), self.wname, self.levels)
        self.levels = self.cudwt.levels
        # Access memory allocated by "pypwt" from pycuda
        self._d_sino = garray.empty(self.sino_shape, np.float32, gpudata=self.cudwt.image_int_ptr())
        self._get_vertical_coeffs()

    def _get_vertical_coeffs(self):
        self._d_vertical_coeffs = {}
        self._d_sino_f = {}
        # Transfer the (0-memset) coefficients in order to get all the shapes
        coeffs = self.cudwt.coeffs
        for i in range(self.cudwt.levels):
            shape = coeffs[i+1][1].shape
            self._d_vertical_coeffs[i+1] = garray.empty(
                shape,
                np.float32,
                gpudata=self.cudwt.coeff_int_ptr(3*i+2)
            )
            self._d_sino_f[i+1] = garray.zeros(
                (shape[0]//2+1, shape[1]),
                dtype=np.complex64
            )


    def _setup_fw_kernel(self):
        self._fw_kernel = CudaKernel(
            "kern_fourierwavelets",
            filename=get_cuda_srcfile("fourier_wavelets.cu"),
            signature="Piif",
        )


    def _destripe_2D(self, d_sino, output):
        # set the "image" for DWT (memcpy D2D)
        self._d_sino.set(d_sino)
        # perform forward DWT
        self.cudwt.forward()
        for i in range(self.cudwt.levels):
            level = i + 1
            d_coeffs = self._d_vertical_coeffs[level]
            d_sino_f = self._d_sino_f[level]
            Ny, Nx = d_coeffs.shape
            # Batched FFT along axis 0
            cufft(
                d_coeffs,
                d_sino_f,
                self._fft_plans[level]["forward"]
            )
            # Dampen the wavelets coefficients
            self._fw_kernel(d_sino_f, Nx, Ny, self.sigma)
            # IFFT
            cuifft(
                d_sino_f,
                d_coeffs,
                self._fft_plans[level]["inverse"]
            )
        # Finally, inverse DWT
        self.cudwt.inverse()
        output.set(self._d_sino)
        return output




from glob import glob
from os import path, getcwd, chdir
from posixpath import join as posix_join
from datetime import datetime
from packaging.version import parse as parse_version
import numpy as np
from h5py import VirtualSource, VirtualLayout
from tomoscan.io import HDF5File
from silx.io.dictdump import dicttoh5
from tifffile import TiffWriter
from .. import version as nabu_version
from ..utils import merged_shape, deprecation_warning, get_available_threads
from ..misc.utils import rescale_data
from .utils import check_h5py_version, convert_dict_values
try:
    from silx.third_party.EdfFile import EdfFile
except ImportError:
    EdfFile = None
try:
    from glymur import Jp2k, set_option as glymur_set_option
    from glymur.version import openjpeg_version, version as glymur_version
    __have_jp2k__ = True
except ImportError:
    __have_jp2k__ = False

def get_datetime():
    """
    Function used by some writers to indicate the current date.
    """
    return datetime.now().replace(microsecond=0).isoformat()


class Writer:
    """
    Base class for all writers.
    """
    def __init__(self, fname):
        self.fname = fname


    def get_filename(self):
        return self.fname



def h5_write_object(h5group, key, value, overwrite=False, default_val=None):
    existing_val = h5group.get(key, default_val)
    if existing_val is not default_val:
        if not overwrite:
            raise OSError("Unable to create link (name already exists): %s" % h5group.name)
        else:
            h5group.pop(key)
    h5group[key] = value


class NXProcessWriter(Writer):
    """
    A class to write Nexus file with a processing result.
    """
    def __init__(self, fname, entry=None, filemode="a", overwrite=False):
        """
        Initialize a NXProcessWriter.

        Parameters
        -----------
        fname: str
            Path to the HDF5 file.
        entry: str, optional
            Entry in the HDF5 file. Default is "entry"
        """
        super().__init__(fname)
        self._set_entry(entry)
        self._filemode = filemode
        self.overwrite = overwrite
        check_h5py_version()


    def _set_entry(self, entry):
        self.entry = entry or "entry"
        data_path = posix_join("/", self.entry)
        self.data_path = data_path


    def write(self, result, process_name, processing_index=0, config=None, data_name="data", is_frames_stack=True):
        """
        Write the result in the current NXProcess group.

        Parameters
        ----------
        result: numpy.ndarray
            Array containing the processing result
        process_name: str
            Name of the processing
        processing_index: int
            Index of the processing (in a pipeline)
        config: dict, optional
            Dictionary containing the configuration.
        """
        swmr = (self._filemode == "r")
        with HDF5File(self.fname, self._filemode, swmr=swmr) as fid:
            nx_entry = fid.require_group(self.data_path)
            if "NX_class" not in nx_entry.attrs:
                nx_entry.attrs["NX_class"] = "NXentry"

            nx_process = nx_entry.require_group(process_name)
            nx_process.attrs['NX_class'] = "NXprocess"

            metadata = {
                "program": "nabu",
                "version": nabu_version,
                "date": get_datetime(),
                "sequence_index": np.int32(processing_index),
            }
            for key, val in metadata.items():
                h5_write_object(nx_process, key, val, overwrite=self.overwrite)

            if config is not None:
                export_dict_to_h5(
                    config,
                    self.fname,
                    posix_join(nx_process.name, "configuration"),
                    overwrite_data=True,
                    mode="a"
                )
                nx_process['configuration'].attrs['NX_class'] = "NXcollection"
            if isinstance(result, dict):
                results_path = posix_join(nx_process.name, "results")
                export_dict_to_h5(
                    result,
                    self.fname,
                    results_path,
                    overwrite_data=self.overwrite,
                    mode="a"
                )
            else:
                nx_data = nx_process.require_group('results')
                results_path = nx_data.name
                nx_data.attrs['NX_class'] = "NXdata"
                nx_data.attrs['signal'] = data_name

                results_data_path = posix_join(results_path, data_name)
                if self.overwrite and results_data_path in fid:
                    del fid[results_data_path]

                if isinstance(result, VirtualLayout):
                    nx_data.create_virtual_dataset(data_name, result)
                else: # assuming array-like
                    nx_data[data_name] = result
                if is_frames_stack:
                    nx_data[data_name].attrs['interpretation'] = "image"
                nx_data.attrs["signal"] = data_name

            # prepare the direct access plots
            nx_process.attrs['default'] = 'results'
            if "default" not in nx_entry.attrs:
                nx_entry.attrs["default"] = posix_join(nx_process.name, 'results')
            # Return the internal path to "results"
            return results_path



def create_virtual_layout(files_or_pattern, h5_path, base_dir=None, axis=0):
    """
    Create a HDF5 virtual layout.

    Parameters
    ----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    """
    prev_cwd = None
    if base_dir is not None:
        prev_cwd = getcwd()
        chdir(base_dir)
    if isinstance(files_or_pattern, str):
        files_list = glob(files_or_pattern)
        files_list.sort()
    else: # list
        files_list = files_or_pattern
    if files_list == []:
        raise ValueError("Nothing found as pattern %s" % files_or_pattern)
    virtual_sources = []
    shapes = []
    for fname in files_list:
        with HDF5File(fname, "r", swmr=True) as fid:
            shape = fid[h5_path].shape
        vsource = VirtualSource(fname, name=h5_path, shape=shape)
        virtual_sources.append(vsource)
        shapes.append(shape)
    total_shape = merged_shape(shapes, axis=axis)
    n_images = total_shape[axis]

    virtual_layout = VirtualLayout(
        shape=total_shape,
        dtype='f'
    )
    start_idx = 0
    for vsource, shape in zip(virtual_sources, shapes):
        n_imgs = shape[axis]
        # Perhaps there is more elegant
        if axis == 0:
            virtual_layout[start_idx:start_idx + n_imgs] = vsource
        elif axis == 1:
            virtual_layout[:, start_idx:start_idx + n_imgs, :] = vsource
        elif axis == 2:
            virtual_layout[:, :, start_idx:start_idx + n_imgs] = vsource
        else:
            raise ValueError("Only axis 0,1,2 are supported")
        #
        start_idx += n_imgs

    if base_dir is not None:
        chdir(prev_cwd)
    return virtual_layout



def merge_hdf5_files(
    files_or_pattern, h5_path, output_file, process_name,
    output_entry=None, output_filemode="a", data_name="data",
    processing_index=0, config=None, base_dir=None,
    axis=0, overwrite=False
):
    """
    Parameters
    -----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    output_file: str
        Path of the output file
    process_name: str
        Name of the process
    output_entry: str, optional
        Output HDF5 root entry (default is "/entry")
    output_filemode: str, optional
        File mode for output file. Default is "a" (append)
    processing_index: int, optional
        Processing index for the output file. Default is 0.
    config: dict, optional
        Dictionary describing the configuration needed to get the results.
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    overwrite: bool, optional
        Whether to overwrite already existing data in the final file.
        Default is False.
    """
    if base_dir is not None:
        prev_cwd = getcwd()
    virtual_layout = create_virtual_layout(files_or_pattern, h5_path, base_dir=base_dir, axis=axis)
    nx_file = NXProcessWriter(
        output_file,
        entry=output_entry, filemode=output_filemode, overwrite=overwrite
    )
    nx_file.write(
        virtual_layout,
        process_name,
        processing_index=processing_index,
        config=config,
        data_name=data_name,
        is_frames_stack=True
    )
    if base_dir is not None and prev_cwd != getcwd():
        chdir(prev_cwd)


class TIFFWriter(Writer):
    def __init__(self, fname, multiframe=False, start_index=0, filemode=None, append=False, big_tiff=None):
        """
        Tiff writer.

        Parameters
        -----------
        fname: str
            Path to the output file name
        multiframe: bool, optional
            Whether to write all data in one single file. Default is False.
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            (unless multiframe is set to True).
            In this case, the output is a series of files `filename_0000.tif`,
            `filename_0001.tif`, etc. This parameter is the starting index for
            file names.
            This option is ignored when multiframe is True.
        filemode: str, optional
            DEPRECATED. Will be ignored. Please refer to 'append'
        append: bool, optional
            Whether to append data to the file rather than overwriting. Default is False.
        big_tiff: bool, optional
            Whether to write in "big tiff" format: https://www.awaresystems.be/imaging/tiff/bigtiff.html
            Default is True when multiframe is True.
            Note that default "standard" tiff cannot exceed 4 GB.

        Notes
        ------
        If multiframe is False (default), then each image will be written in a
        dedicated tiff file.
        """
        super().__init__(fname)
        self.multiframe = multiframe
        self.start_index = start_index
        self.append = append
        if big_tiff is None:
            big_tiff = multiframe
        if multiframe and not big_tiff:
            # raise error ?
            print("big_tiff was set to False while multiframe was set to True. This will probably be problematic.")
        self.big_tiff = big_tiff
        # Compat.
        self.filemode = filemode
        if filemode is not None:
            deprecation_warning("Ignored parameter 'filemode'. Please use the 'append' parameter")
        #

    def _write_tiff(self, data, config=None, filename=None):
        # TODO metadata
        filename = filename or self.fname
        with TiffWriter(filename, bigtiff=self.big_tiff, append=self.append) as tif:
            tif.write(data)

    def write(self, data, *args, config=None, **kwargs):
        # Single image, or multiple image in the same file
        if self.multiframe or data.ndim < 3 or ((data.ndim == 3) and (data.shape[0] == 1)):
            self._write_tiff(data, config=config)
        # Multiple image, one file per image
        else:
            dirname, rel_filename = path.split(self.fname)
            prefix, ext = path.splitext(rel_filename)
            for i in range(data.shape[0]):
                curr_rel_filename = prefix + str("_%04d" % (self.start_index + i)) + ext
                fname = path.join(dirname, curr_rel_filename)
                self._write_tiff(data[i], filename=fname, config=config)

    def get_filename(self):
        if self.multiframe:
            return self.fname
        else:
            return path.dirname(self.fname)


class EDFWriter(Writer):
    def __init__(self, fname, start_index=0, filemode="w"):
        """
        EDF (ESRF Data Format) writer.

        Parameters
        -----------
        fname: str
            Path to the output file name
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            In this case, the output is a series of files `filename_0000.tif`,
            `filename_0001.edf`, etc. This parameter is the starting index for
            file names.
        """
        super().__init__(fname)
        self.filemode = filemode
        self.start_index = start_index

    def _write_edf(self, data, config=None, filename=None, filemode=None):
        if filename is None:
            filename = self.fname
        edf = EdfFile(filename, access="w+b")
        config = config or {}
        config["software"] = str("nabu %s" % nabu_version)
        config["date"] = get_datetime()
        edf.WriteImage(
            config,
            data
        )
        edf = None


    def write(self, data, *args, config=None, **kwargs):
        if data.ndim < 3:
            self._write_edf(data, config=config)
            return
        if (data.ndim == 3) and (data.shape[0] == 1):
            self._write_edf(data[0], config=config)
            return
        dirname, rel_filename = path.split(self.fname)
        prefix, ext = path.splitext(rel_filename)
        for i in range(data.shape[0]):
            curr_rel_filename = prefix + str("_%04d" % (self.start_index + i)) + ext
            fname = path.join(dirname, curr_rel_filename)
            self._write_edf(data[i], filename=fname, config=config)

    def get_filename(self):
        return path.dirname(self.fname)


class JP2Writer(Writer):
    def __init__(
        self, fname, start_index=0, filemode="wb",
        psnr=None, cratios=None, auto_convert=True, float_clip_values=None, n_threads=None
    ):
        """
        JPEG2000 writer. This class requires the python package `glymur` and the
        library `libopenjp2`.

        Parameters
        -----------
        fname: str
            Path to the output file name
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            The output is a series of files `filename_0000.tif`, `filename_0001.tif`, etc.
            This parameter is the starting index for file names.
        psnr: list of int, optional
            The PSNR (Peak Signal-to-Noise ratio) for each jpeg2000 layer.
            This defines a quality metric for lossy compression.
            The number "0" stands for lossless compression.
        cratios: list of int, optional
            Compression ratio for each jpeg2000 layer
        auto_convert: bool, optional
            Whether to automatically cast floating point data to uint16.
            Default is True.
        float_clip_values: tuple of floats, optional
            If set to a tuple of two values (min, max), then each image values will be clipped
            to these minimum and maximum values.
        n_threads: int, optional
            Number of threads to use for encoding. Default is the number of available threads.
            Needs libopenjpeg >= 2.4.0.
        """
        super().__init__(fname)
        if not(__have_jp2k__):
            raise ValueError("Need glymur python package and libopenjp2 library")
        self.setup_multithread_encoding(n_threads=n_threads, what_if_not_available="ignore")
        self.filemode = filemode
        self.start_index = start_index
        self.auto_convert = auto_convert
        if psnr is not None and np.isscalar(psnr):
            psnr = [psnr]
        self.psnr = psnr
        self.cratios = cratios
        self._vmin = None
        self._vmax = None
        self.clip_float = False
        if float_clip_values is not None:
            self._float_clip_min, self._float_clip_max = float_clip_values
            self.clip_float = True


    @staticmethod
    def setup_multithread_encoding(n_threads=None, what_if_not_available="ignore"):
        """
        Setup OpenJpeg multi-threaded encoding.

        Parameters
        -----------
        n_threads: int, optional
            Number of threads. If not provided, all available threads are used.
        na: str, optional
            What to do if requirements are not fulfilled. Possible values are:
               - "ignore": do nothing, proceed
               - "print": show an information message
               - "raise": raise an error
        """
        required_glymur_version = "0.9.3"
        required_openjpeg_version = "2.4.0"

        def not_available(msg):
            if what_if_not_available == "raise":
                raise ValueError(msg)
            elif what_if_not_available == "print":
                print(msg)

        requirements = [
            ("glymur", glymur_version, required_glymur_version),
            ("libopenjpeg", openjpeg_version, required_openjpeg_version)
        ]
        for what, current_version, required_version in requirements:
            if parse_version(current_version) < parse_version(required_version):
                not_available(
                    "%s version >= %s is required for multi-threaded encoding (current version: %s)"
                    % (what, required_version, current_version)
                )
                return
        if n_threads is None:
            n_threads = get_available_threads()
        glymur_set_option("lib.num_threads", n_threads)


    def _write_jp2k(self, data, filename=None):
        if filename is None:
            filename = self.fname
        if data.dtype != np.uint16 and self.auto_convert:
            if self.clip_float:
                data = np.clip(data, self._float_clip_min, self._float_clip_max)
            data = rescale_data(data, 0, 65535, data_min=self._vmin, data_max=self._vmax)
            data = data.astype(np.uint16)
        #
        jp2 = Jp2k(filename, data=data, psnr=self.psnr, cratios=self.cratios)


    def write(self, data, *args, **kwargs):
        if data.ndim < 3:
            self._write_jp2k(data)
            return
        if (data.ndim == 3) and (data.shape[0] == 1):
            self._write_jp2k(data[0])
            return
        dirname, rel_filename = path.split(self.fname)
        prefix, ext = path.splitext(rel_filename)
        for i in range(data.shape[0]):
            curr_rel_filename = prefix + str("_%04d" % (self.start_index + i)) + ext
            fname = path.join(dirname, curr_rel_filename)
            self._write_jp2k(data[i], filename=fname)


    def get_filename(self):
        return path.dirname(self.fname)


class NPYWriter(Writer):
    def __init__(self,  fname):
        super().__init__(fname)

    def write(self, result, *args, **kwargs):
        np.save(self.fname, result)


class NPZWriter(Writer):
    def __init__(self,  fname):
        super().__init__(fname)

    def write(self, result, *args, **kwargs):
        save_args = {"result": result}
        config = kwargs.get("config", None)
        if config is not None:
            save_args["configuration"] = config
        np.savez(self.fname, **save_args)


Writers = {
    "h5": NXProcessWriter,
    "hdf5": NXProcessWriter,
    "nx": NXProcessWriter,
    "nexus": NXProcessWriter,
    "npy": NPYWriter,
    "npz": NPZWriter,
    "tif": TIFFWriter,
    "tiff": TIFFWriter,
    "j2k": JP2Writer,
    "jp2": JP2Writer,
    "jp2k": JP2Writer,
    "edf": EDFWriter,
}


def export_dict_to_h5(dic, h5file, h5path, overwrite_data=True, mode="a"):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing None with "None"

    Parameters
    -----------
    dic: dict
        Dictionary containing the options
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    overwrite_data: bool, optional
        Whether to overwrite data when writing HDF5. Default is True
    mode: str, optional
        File mode. Default is "a" (append).
    """
    modified_dic = convert_dict_values(
        dic,
        {None: "None"},
    )
    update_mode = {True: 'modify', False: 'add'}[bool(overwrite_data)]
    return dicttoh5(
        modified_dic,
        h5file=h5file,
        h5path=h5path,
        update_mode=update_mode,
        mode=mode
    )


from .alignment import AlignmentBase
from .cor import CenterOfRotation, CenterOfRotationSlidingWindow, CenterOfRotationGrowingWindow, CenterOfRotationAdaptiveSearch
from .cor_sino import SinoCor
from .distortion import estimate_flat_distortion
from .focus import CameraFocus
from .tilt import CameraTilt
from .translation import DetectorTranslationAlongBeam



import numpy as np
import scipy.signal

def correct_spikes(image, threshold):
    """
    Perform a conditional median filtering
    The filtering is done in-place, meaning that the array content is modified.

    Parameters
    ----------
    image: numpy.ndarray
        Image to filter
    threshold: float
        Median filter threshold
    """
    m_im = scipy.signal.medfilt2d(image)
    fixed_part = np.array(image[[0, 0, -1, -1], [0, -1, 0, -1]])

    where = abs(image - m_im) > threshold
    image[where] = m_im[where]
    image[[0, 0, -1, -1], [0, -1, 0, -1]] = fixed_part

    return image



import math
import numpy as np

def pad_interpolate(im, padded_img_shape_vh, translation_vh=None, padding_mode="reflect"):
    """
    This function produces a centered padded image and , optionally if translation_vh is set,
    performs a Fourier shift of the whole image. In case of translation, the image is first padded to
    a larger extent, that encompasses the final padded with plus a translation margin, and then translated,
    and final recut to the required padded width.
    The values are translated: if a feature appear at x in the original image it will appear at pad+translation+x
    in the final image.

    Parameters
    ------------
    im: np.ndaray
       the input image
    translation_vh: a sequence of two float
       the vertical and horizontal shifts
    """

    if translation_vh is not None:
        pad_extra = 2 ** (1 + np.ceil(np.log2(np.maximum(1, np.ceil(abs(np.array(translation_vh)))))).astype(np.int32))
    else:
        pad_extra = [0, 0]

    origy, origx = im.shape
    rety = padded_img_shape_vh[0] + pad_extra[0]
    retx = padded_img_shape_vh[1] + pad_extra[1]

    xpad = [0, 0]
    xpad[0] = math.ceil((retx - origx) / 2)
    xpad[1] = retx - origx - xpad[0]

    ypad = [0, 0]
    ypad[0] = math.ceil((rety - origy) / 2)
    ypad[1] = rety - origy - ypad[0]

    y2 = origy - ypad[1]
    x2 = origx - xpad[1]

    if ypad[0] + 1 > origy or xpad[0] + 1 > origx or y2 < 1 or x2 < 1:
        raise ValueError("Too large padding for this reflect padding type")

    padded_im = np.pad(im, pad_width=((ypad[0], ypad[1]), (xpad[0], xpad[1])), mode=padding_mode)

    if translation_vh is not None:
        freqs_list = list(map(np.fft.fftfreq, padded_im.shape))
        shifts_list = [np.exp(-2.0j * np.pi * freqs * trans) for (freqs, trans) in zip(freqs_list, translation_vh)]
        shifts_2D = shifts_list[0][:, None] * shifts_list[1][None, :]
        padded_im = np.fft.ifft2(np.fft.fft2(padded_im) * shifts_2D).real
        padded_im = recut(padded_im, padded_img_shape_vh)

    return padded_im


def recut(im, new_shape_vh):
    """
    This method implements a centered cut which reverts the centered padding
    applied in the present class.

    Parameters
    -----------
    im: np.ndarray
        A 2D image.
    new_shape_vh:  tuple
        The shape of the cutted image.

    Returns
    --------
        The image cutted to new_shape_vh.
    """

    new_shape_vh = np.array(new_shape_vh)

    old_shape_vh = np.array(im.shape)

    center_vh = (old_shape_vh - 1) / 2

    start_vh = np.round(0.5 + center_vh - new_shape_vh / 2).astype(np.int32)
    end_vh = start_vh + new_shape_vh

    return im[start_vh[0] : end_vh[0], start_vh[1] : end_vh[1]]



import numpy as np
from ..utils import check_supported, deprecated_class
from scipy.ndimage.filters import median_filter

# COMPAT.
from .flatfield import FlatField as Flatfield_, FlatFieldArrays as FlatFieldArrays_, FlatFieldDataUrls as FlatFieldDataUrls_
FlatField = deprecated_class(
    "preproc.ccd.FlatField was moved to preproc.flatfield.FlatField", do_print=True
)(Flatfield_)
FlatFieldArrays = deprecated_class(
    "preproc.ccd.FlatFieldArrays was moved to preproc.flatfield.FlatFieldArrays", do_print=True
)(FlatFieldArrays_)
FlatFieldDataUrls = deprecated_class(
    "preproc.ccd.FlatFieldDataUrls was moved to preproc.flatfield.FlatFieldDataUrls", do_print=True
)(FlatFieldDataUrls_)
#


class CCDFilter:
    """
    A base class for processings applied on radios.
    It is designed to handle chunks of radios (see the documentation for more
    information on the radios chunk concept).
    """

    _supported_ccd_corrections = ["median_clip"]

    def __init__(
        self,
        radios_shape: tuple,
        correction_type: str = "median_clip",
        median_clip_thresh: float = 0.1,
        abs_diff=False,
        preserve_borders=False,
    ):
        """
        Initialize a CCDCorrection instance.

        Parameters
        -----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        correction_type: str
            Correction type for radios ("median_clip", "sigma_clip", ...)
        median_clip_thresh: float, optional
            Threshold for the median clipping method.
        abs_diff: boolean
            by default False:  the correction is triggered  when img - median > threshold.
            If equals True:    correction is triggered for abs(img-media) > threshold
        preserve borders: boolean
            by default False:
            If equals True:  the borders (width=1) are not modified.


        Notes
        ------
        A CCD correction is a process (usually filtering) taking place in the
        radios space.
        Available filters:
           - median_clip: if the value of the current pixel exceeds the median
             of adjacent pixels (a 3x3 neighborhood) more than a threshold,
             then this pixel value is set to the median value.
        """
        self._set_radios_shape(radios_shape)
        check_supported(correction_type, self._supported_ccd_corrections, "CCD correction mode")
        self.correction_type = correction_type
        self.median_clip_thresh = median_clip_thresh
        self.abs_diff = abs_diff
        self.preserve_borders = preserve_borders


    def _set_radios_shape(self, radios_shape):
        if len(radios_shape) == 2:
            self.radios_shape = (1,) + radios_shape
        elif len(radios_shape) == 3:
            self.radios_shape = radios_shape
        else:
            raise ValueError("Expected radios to have 2 or 3 dimensions")
        n_radios, n_z, n_x = self.radios_shape
        self.n_radios = n_radios
        self.n_angles = n_radios
        self.shape = (n_z, n_x)

    @staticmethod
    def median_filter(img):
        """
        Perform a median filtering on an image.
        """
        return median_filter(img, (3, 3), mode="reflect")

    def median_clip_mask(self, img, return_medians=False):
        """
        Compute a mask indicating whether a pixel is valid or not, according to
        the median-clip method.

        Parameters
        ----------
        img: numpy.ndarray
            Input image
        return_medians: bool, optional
            Whether to return the median values additionally to the mask.
        """
        median_values = self.median_filter(img)
        if not self.abs_diff:
            invalid_mask = img >= median_values + self.median_clip_thresh
        else:
            invalid_mask = abs(img - median_values) > self.median_clip_thresh

        if return_medians:
            return invalid_mask, median_values
        else:
            return invalid_mask

    def median_clip_correction(self, radio, output=None):
        """
        Compute the median clip correction on one image.

        Parameters
        ----------
        radios: numpy.ndarray, optional
            A radio image.
        output: numpy.ndarray, optional
            Output array
        """
        assert radio.shape == self.shape
        if output is None:
            output = np.copy(radio)
        else:
            output[:] = radio[:]
        invalid_mask, medians = self.median_clip_mask(radio, return_medians=True)

        if self.preserve_borders:
            fixed_border = np.array(radio[[0, 0, -1, -1], [0, -1, 0, -1]])

        output[invalid_mask] = medians[invalid_mask]

        if self.preserve_borders:
            output[[0, 0, -1, -1], [0, -1, 0, -1]] = fixed_border

        return output


# Compat.
CCDCorrection = deprecated_class("CCDCorrection is replaced with CCDFilter", do_print=True)(CCDFilter)


class Log:
    """
    Helper class to take -log(radios)

    Parameters
    -----------
    clip_min: float, optional
        Before taking the logarithm, the values are clipped to this minimum.
    clip_max: float, optional
        Before taking the logarithm, the values are clipped to this maximum.
    """

    def __init__(self, radios_shape, clip_min=None, clip_max=None):
        self.radios_shape = radios_shape
        self.clip_min = clip_min
        self.clip_max = clip_max

    def take_logarithm(self, radios):
        """
        Take the negative logarithm of a radios chunk.
        Processing is done in-place !

        Parameters
        -----------
        radios: array
            Radios chunk.
        """
        if (self.clip_min is not None) or (self.clip_max is not None):
            np.clip(radios, self.clip_min, self.clip_max, out=radios)
            np.log(radios, out=radios)
        else:
            np.log(radios, out=radios)
        radios[:] *= -1
        return radios



import numpy as np
import pytest
from nabu.misc.binning import *

@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.data = np.arange(100*99, dtype=np.uint16).reshape(100, 99)
    cls.tol = 1e-5

#
# Alternate implementation of binning
# TODO do we have off-the shelf binning in scipy or skimage ?
#

def binning2_ref(img):
    return img.reshape(img.shape[0] //2, 2, img.shape[1]//2, 2).mean(axis=-1).mean(axis=1)

def binning2_horiz_ref(img):
    return img.reshape(img.shape[0], img.shape[1]//2, 2).mean(axis=-1)

def binning2_vertic_ref(img):
    return img.reshape(img.shape[0] //2, 2, img.shape[1]).mean(axis=1)

def binning3_ref(img):
    return img.reshape(img.shape[0] //3, 3, img.shape[1]//3, 3).mean(axis=-1).mean(axis=1)

def binning3_horiz_ref(img):
    return img.reshape(img.shape[0], img.shape[1]//3, 3).mean(axis=-1)

def binning3_vertic_ref(img):
    return img.reshape(img.shape[0]//3, 3, img.shape[1]).mean(axis=1)


@pytest.mark.usefixtures('bootstrap')
class TestUnsharp:

    def checkResult(self, res, ref, name):
        mae = np.max(np.abs(res - ref))
        assert mae < self.tol, "%s: max error is too high (%.2e > %.2e)" % (name, mae, self.tol)


    def testBinning2(self):
        data = self.data
        res = binning2(data)
        ref = binning2_ref(data[:, :-1])
        self.checkResult(res, ref, "binning 2x2")

        res = binning2_horiz(data)
        ref = binning2_horiz_ref(data[:, :-1])
        self.checkResult(res, ref, "horizontal 2-binning")

        res = binning2_vertic(data)
        ref = binning2_vertic_ref(data)
        self.checkResult(res, ref, "horizontal 2-binning")


    def testBinning3(self):
        data = self.data
        res = binning3(data)
        ref = binning3_ref(data[:-1, :])
        self.checkResult(res, ref, "binning 3x3")

        res = binning3_horiz(data)
        ref = binning3_horiz_ref(data)
        self.checkResult(res, ref, "horizontal 3-binning")

        res = binning3_vertic(data)
        ref = binning3_vertic_ref(data[:-1, :])
        self.checkResult(res, ref, "horizontal 3-binning")




import pytest
import numpy as np
from scipy.misc import ascent
from scipy.ndimage.filters import median_filter
from nabu.testutils import generate_tests_scenarios
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
if __has_pycuda__:
    from nabu.cuda.medfilt import MedianFilter
    import pycuda.gpuarray as garray


scenarios = generate_tests_scenarios({
    "input_on_gpu": [False, True],
    "output_on_gpu": [False, True],
    "footprint": [(3, 3), (5, 5)],
    "mode": ["reflect", "nearest", "wrap"],
    "batched_2d": [False, True],
})


@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls

    cls.data = np.ascontiguousarray(ascent()[::2, ::2][:-1, :], dtype="f")
    cls.tol = 1e-7
    cls.ctx = get_cuda_context()
    cls.allocate_numpy_arrays()
    cls.allocate_cuda_arrays()


@pytest.mark.skipif(
    not(__has_pycuda__),
    reason="Need Cuda/pycuda for this test"
)
@pytest.mark.usefixtures('bootstrap')
class TestMedianFilter(object):

    @classmethod
    def allocate_numpy_arrays(cls):
        shape = cls.data.shape
        cls.input = cls.data
        cls.input3d = np.tile(cls.input, (2, 1, 1))

    @classmethod
    def allocate_cuda_arrays(cls):
        shape = cls.data.shape
        cls.d_input = garray.to_gpu(cls.input)
        cls.d_output = garray.zeros_like(cls.d_input)
        cls.d_input3d = garray.to_gpu(cls.input3d)
        cls.d_output3d = garray.zeros_like(cls.d_input3d)

    # parametrize on a class method will use the same class, and launch this
    # method with different scenarios.
    @pytest.mark.parametrize("config", scenarios)
    def testMedfilt(self, config):
        if config["input_on_gpu"]:
            input_data = self.d_input if not(config["batched_2d"]) else self.d_input3d
        else:
            input_data = self.input if not(config["batched_2d"]) else self.input3d
        if config["output_on_gpu"]:
            output_data = self.d_output if not(config["batched_2d"] ) else self.d_output3d
        else:
            output_data =  None
        # Cuda median filter
        medfilt = MedianFilter(
            input_data.shape,
            footprint=config["footprint"],
            mode=config["mode"],
            cuda_options={"ctx": self.ctx},
        )
        res = medfilt.medfilt2(
            input_data,
            output=output_data
        )
        if config["output_on_gpu"]:
            res = res.get()
        # Reference (scipy)
        ref = median_filter(
            self.input,
            config["footprint"][0],
            mode=config["mode"]
        )
        max_absolute_error = np.max(np.abs(res - ref))
        assert max_absolute_error < self.tol, "Something wrong with configuration %s" % str(config)



import math
import numpy as np
from ..misc import fourier_filters
from .alignment import AlignmentBase, plt, progress_bar, local_fftn, local_ifftn

# three possible  values for the validity check, which can optionally be returned by the find_shifts methods
cor_result_validity =  {
    "unknown": "unknown",
    "sound": "sound",
    "correct": "sound",
    "questionable": "questionable",
}


class CenterOfRotation(AlignmentBase):
    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        shift_axis: int = -1,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
    ):
        """Find the Center of Rotation (CoR), given two images.

        This method finds the half-shift between two opposite images, by
        means of correlation computed in Fourier space.

        The output of this function, allows to compute motor movements for
        aligning the sample rotation axis. Given the following values:

        - L1: distance from source to motor
        - L2: distance from source to detector
        - ps: physical pixel size
        - v: output of this function

        displacement of motor = (L1 / L2 * ps) * v

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        shift_axis: int
            Axis along which we want the shift to be computed. Default is -1 (horizontal).
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`
        return_validity: a boolean, defaults to false
            if set to True adds a second return value which may have three string values.
            These values are "unknown", "sound", "questionable".
            It will be "uknown" if the  validation method is not implemented
            and it will be "sound" or "questionable" if it is implemented.


        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        float
            Estimated center of rotation position from the center of the RoI in pixels.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... CoR_calc = CenterOfRotation()
        ... cor_position = CoR_calc.find_shift(radio1, radio2)

        Or for noisy images:

        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3))
        """

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        cc = self._compute_correlation_fft(img_1, img_2, padding_mode, high_pass=high_pass, low_pass=low_pass)
        img_shape = img_2.shape
        cc_vs = np.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
        cc_hs = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

        (f_vals, fv, fh) = self.extract_peak_region_2d(cc, peak_radius=peak_fit_radius, cc_vs=cc_vs, cc_hs=cc_hs)
        fitted_shifts_vh = self.refine_max_position_2d(f_vals, fv, fh)

        if return_validity:
            return fitted_shifts_vh[shift_axis] / 2.0, cor_result_validity["unknown"]
        else:
            return fitted_shifts_vh[shift_axis] / 2.0


class CenterOfRotationSlidingWindow(CenterOfRotation):
    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        side,
        window_width=None,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
    ):
        """Semi-automatically find the Center of Rotation (CoR), given two images
        or sinograms. Suitable for half-aquisition scan.

        This method finds the half-shift between two opposite images,  by
        minimizing difference over a moving window.

        The output of this function, allows to compute motor movements for
        aligning the sample rotation axis. Given the following values:

        - L1: distance from source to motor
        - L2: distance from source to detector
        - ps: physical pixel size
        - v: output of this function

        displacement of motor = (L1 / L2 * ps) * v

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        side: string
            Expected region of the CoR. Allowed values: 'left', 'center' or 'right'.
        window_width: int, optional
            Width of window that will slide on the other image / part of the
            sinogram. Default is None.
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`
        return_validity: a boolean, defaults to false
            if set to True adds a second return value which may have three string values.
            These values are "unknown", "sound", "questionable".
            It will be "uknown" if the  validation method is not implemented
            and it will be "sound" or "questionable" if it is implemented.

        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        float
            Estimated center of rotation position from the center of the RoI in pixels.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... CoR_calc = CenterOfRotationSlidingWindow()
        ... cor_position = CoR_calc.find_shift(radio1, radio2)

        Or for noisy images:

        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3))
        """

        validity_check_result = cor_result_validity["unknown"]

        if side is None:
            raise ValueError("Side should be one of 'left', 'right', or 'center'. 'None' was given instead")

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning(
                "Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius
            )
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(
            img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_2 = self._prepare_image(
            img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_shape = img_2.shape

        if window_width is None:
            if side.lower() == "center":
                window_width = round(img_shape[-1] / 4.0 * 3.0)
            else:
                window_width = round(img_shape[-1] / 10)
        window_shift = window_width // 2
        window_width = window_shift * 2 + 1

        if side.lower() == "right":
            win_2_start = 0
        elif side.lower() == "left":
            win_2_start = img_shape[-1] - window_width
        elif side.lower() == "center":
            win_2_start = img_shape[-1] // 2 - window_shift
        else:
            raise ValueError("Side should be one of 'left', 'right', or 'center'. '%s' was given instead" % side.lower())

        win_2_end = win_2_start + window_width

        # number of pixels where the window will "slide".
        n = img_shape[-1] - window_width
        diffs_mean = np.zeros((n,), dtype=img_1.dtype)
        diffs_std = np.zeros((n,), dtype=img_1.dtype)

        for ii in progress_bar(range(n), verbose=self.verbose):
            win_1_start, win_1_end = ii, ii + window_width
            img_diff = img_1[:, win_1_start:win_1_end] - img_2[:, win_2_start:win_2_end]
            diffs_abs = np.abs(img_diff)
            diffs_mean[ii] = diffs_abs.mean()
            diffs_std[ii] = diffs_abs.std()

        diffs_mean = diffs_mean.min() - diffs_mean
        win_ind_max = np.argmax(diffs_mean)

        diffs_std = diffs_std.min() - diffs_std
        if not win_ind_max == np.argmax(diffs_std):
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]
        else:
            validity_check_result = cor_result_validity["sound"]

        (f_vals, f_pos) = self.extract_peak_regions_1d(diffs_mean, peak_radius=peak_fit_radius)
        win_pos_max, win_val_max = self.refine_max_position_1d(f_vals, return_vertex_val=True)

        cor_h = -(win_2_start - (win_ind_max + win_pos_max)) / 2.0

        if (side.lower() == "right" and win_ind_max == 0) or (side.lower() == "left" and win_ind_max == n):
            self.logger.warning("Sliding window width %d might be too large!" % window_width)

        if self.verbose:
            cor_pos = -(win_2_start - np.arange(n)) / 2.0

            print("Lowest difference window: index=%d, range=[0, %d]" % (win_ind_max, n))
            print("CoR tested for='%s', found at voxel=%g (from center)" % (side, cor_h))

            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.stem(cor_pos, diffs_mean, label="Mean difference")
            ax.stem(cor_h, win_val_max, linefmt="C1-", markerfmt="C1o", label="Best mean difference")
            ax.stem(cor_pos, -diffs_std, linefmt="C2-", markerfmt="C2o", label="Std-dev difference")
            ax.set_title("Window dispersions")
            plt.show(block=False)

        if return_validity:
            return cor_h, validity_check_result
        else:
            return cor_h


class CenterOfRotationGrowingWindow(CenterOfRotation):
    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        side="all",
        min_window_width=11,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_validity=False,
    ):
        """Automatically find the Center of Rotation (CoR), given two images or
        sinograms. Suitable for half-aquisition scan.

        This method finds the half-shift between two opposite images,  by
        minimizing difference over a moving window.

        The output of this function, allows to compute motor movements for
        aligning the sample rotation axis. Given the following values:

        - L1: distance from source to motor
        - L2: distance from source to detector
        - ps: physical pixel size
        - v: output of this function

        displacement of motor = (L1 / L2 * ps) * v

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        side: string, optional
            Expected region of the CoR. Allowed values: 'left', 'center', 'right', or 'all'.
            Default is 'all'.
        min_window_width: int, optional
            Minimum window width that covers the common region of the two images /
            sinograms. Default is 11.
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`
        return_validity: a boolean, defaults to false
            if set to True adds a second return value which may have three string values.
            These values are "unknown", "sound", "questionable".
            It will be "uknown" if the  validation method is not implemented
            and it will be "sound" or "questionable" if it is implemented.



        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        float
            Estimated center of rotation position from the center of the RoI in pixels.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... CoR_calc = CenterOfRotationGrowingWindow()
        ... cor_position = CoR_calc.find_shift(radio1, radio2)

        Or for noisy images:

        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3))
        """

        validity_check_result = cor_result_validity["unknown"]

        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(
            img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_2 = self._prepare_image(
            img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, high_pass=high_pass, low_pass=low_pass
        )
        img_shape = img_2.shape

        def window_bounds(mid_point, window_max_width=img_shape[-1]):
            return (
                np.fmax(np.ceil(mid_point - window_max_width / 2), 0).astype(np.intp),
                np.fmin(np.ceil(mid_point + window_max_width / 2), img_shape[-1]).astype(np.intp),
            )

        img_lower_half_size = np.floor(img_shape[-1] / 2).astype(np.intp)
        img_upper_half_size = np.ceil(img_shape[-1] / 2).astype(np.intp)

        if side.lower() == "right":
            win_1_mid_start = img_lower_half_size
            win_1_mid_end = np.floor(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
            win_2_mid_start = -img_upper_half_size + min_window_width
            win_2_mid_end = img_upper_half_size
        elif side.lower() == "left":
            win_1_mid_start = -img_lower_half_size + min_window_width
            win_1_mid_end = img_lower_half_size
            win_2_mid_start = img_upper_half_size
            win_2_mid_end = np.ceil(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
        elif side.lower() == "center":
            win_1_mid_start = 0
            win_1_mid_end = img_shape[-1]
            win_2_mid_start = 0
            win_2_mid_end = img_shape[-1]
        elif side.lower() == "all":
            win_1_mid_start = -img_lower_half_size + min_window_width
            win_1_mid_end = np.floor(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
            win_2_mid_start = -img_upper_half_size + min_window_width
            win_2_mid_end = np.ceil(img_shape[-1] * 3 / 2).astype(np.intp) - min_window_width
        else:
            raise ValueError("Side should be one of 'left', 'right', or 'center'. '%s' was given instead" % side.lower())

        n1 = win_1_mid_end - win_1_mid_start
        n2 = win_2_mid_end - win_2_mid_start

        if not n1 == n2:
            raise ValueError(
                "Internal error: the number of window steps for the two images should be the same."
                + "Found the following configuration instead => Side: %s, #1: %d, #2: %d" % (side, n1, n2)
            )

        diffs_mean = np.zeros((n1,), dtype=img_1.dtype)
        diffs_std = np.zeros((n1,), dtype=img_1.dtype)

        for ii in progress_bar(range(n1), verbose=self.verbose):
            win_1 = window_bounds(win_1_mid_start + ii)
            win_2 = window_bounds(win_2_mid_end - ii)
            img_diff = img_1[:, win_1[0] : win_1[1]] - img_2[:, win_2[0] : win_2[1]]
            diffs_abs = np.abs(img_diff)
            diffs_mean[ii] = diffs_abs.mean()
            diffs_std[ii] = diffs_abs.std()

        diffs_mean = diffs_mean.min() - diffs_mean
        win_ind_max = np.argmax(diffs_mean)

        diffs_std = diffs_std.min() - diffs_std
        if not win_ind_max == np.argmax(diffs_std):
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]
        else:
            validity_check_result = cor_result_validity["sound"]

        (f_vals, f_pos) = self.extract_peak_regions_1d(diffs_mean, peak_radius=peak_fit_radius)
        win_pos_max, win_val_max = self.refine_max_position_1d(f_vals, return_vertex_val=True)

        cor_h = (win_1_mid_start + (win_ind_max + win_pos_max) - img_upper_half_size) / 2.0

        if (side.lower() == "right" and win_ind_max == 0) or (side.lower() == "left" and win_ind_max == n1):
            self.logger.warning("Minimum growing window width %d might be too large!" % min_window_width)

        if self.verbose:
            cor_pos = (win_1_mid_start + np.arange(n1) - img_upper_half_size) / 2.0

            self.logger.info("Lowest difference window: index=%d, range=[0, %d]" % (win_ind_max, n1))
            self.logger.info("CoR tested for='%s', found at voxel=%g (from center)" % (side, cor_h))

            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.stem(cor_pos, diffs_mean, label="Mean difference")
            ax.stem(cor_h, win_val_max, linefmt="C1-", markerfmt="C1o", label="Best mean difference")
            ax.stem(cor_pos, -diffs_std, linefmt="C2-", markerfmt="C2o", label="Std-dev difference")
            ax.set_title("Window dispersions")
            plt.show(block=False)

        if return_validity:
            return cor_h, validity_check_result
        else:
            return cor_h


class CenterOfRotationAdaptiveSearch(CenterOfRotation):
    """This adaptive method works by applying a gaussian which highlights, by apodisation, a region
    which can possibly contain the good center of rotation.
    The whole image is spanned during several applications of the apodisation. At each application
    the apodisation function, which is a gaussian, is moved to a new guess position.
    The lenght of the step, by which the gaussian is moved, and its sigma are
    obtained by multiplying the shortest distance from the left or right border with
    a self.step_fraction and  self.sigma_fraction factors which ensure global overlapping.
    for each step a region around the CoR  of each image is selected, and the regions of the two images
    are compared to  calculate a cost function. The value of the cost function, at its minimum
    is used to select the best step at which the CoR is taken as final result.
    The option filtered_cost= True (default) triggers the filtering (according to low_pass and high_pass)
    of the two images which are used for he cost function. ( Note: the low_pass and high_pass options
    are used, if given, also without the filtered_cost option, by being passed to the base class
    CenterOfRotation )
    """

    sigma_fraction = 1.0 / 4.0
    step_fraction = 1.0 / 6.0

    def find_shift(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        high_pass=None,
        low_pass=None,
        margins=None,
        filtered_cost=True,
        return_validity=False,
    ):
        """Find the Center of Rotation (CoR), given two images.

        This method finds the half-shift between two opposite images, by
        means of correlation computed in Fourier space.
        A global search is done on on the detector span (minus a margin) without assuming centered scan conditions.

        The output of this function, allows to compute motor movements for
        aligning the sample rotation axis. Given the following values:

        - L1: distance from source to motor
        - L2: distance from source to detector
        - ps: physical pixel size
        - v: output of this function

        displacement of motor = (L1 / L2 * ps) * v

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        low_pass: float or sequence of two floats.
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`.
        margins:  None or a couple of floats or ints
            if margins is None or in the form of  (margin1,margin2) the search is done between margin1 and  dim_x-1-margin2.
            If left to None then by default (margin1,margin2)  = ( 10, 10 ).
        filtered_cost: boolean.
            True by default. It triggers the use of filtered images in the calculation of the cost function.
        return_validity: a boolean, defaults to false
            if set to True adds a second return value which may have three string values.
            These values are "unknown", "sound", "questionable".
            It will be "uknown" if the  validation method is not implemented
            and it will be "sound" or "questionable" if it is implemented.

        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        float
            Estimated center of rotation position from the center of the RoI in pixels.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... CoR_calc = CenterOfRotationAdaptiveSearch()
        ... cor_position = CoR_calc.find_shift(radio1, radio2)

        Or for noisy images:

        >>> cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3), high_pass=20, low_pass=1   )
        """

        validity_check_result = cor_result_validity["unknown"]

        self._check_img_pair_sizes(img_1, img_2)

        used_type = img_1.dtype

        roi_yxhw = self._determine_roi(img_1.shape, roi_yxhw)

        if filtered_cost and (low_pass is not None or high_pass is not None):
            img_filter = fourier_filters.get_bandpass_filter(
                img_1.shape[-2:],
                cutoff_lowpass=low_pass,
                cutoff_highpass=high_pass,
                use_rfft=True,
                data_type=self.data_type,
            )
            # fft2 and iff2 use axes=(-2, -1) by default
            img_filtered_1 = local_ifftn(local_fftn(img_1, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
            img_filtered_2 = local_ifftn(local_fftn(img_2, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real
        else:
            img_filtered_1 = img_1
            img_filtered_2 = img_2

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        img_filtered_1 = self._prepare_image(img_filtered_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_filtered_2 = self._prepare_image(img_filtered_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        dim_radio = img_1.shape[1]

        if margins is None:
            lim_1, lim_2 = 10, dim_radio - 1 - 10
        else:
            lim_1, lim_2 = margins
            lim_2 = dim_radio - 1 - lim_2

        if lim_1 < 1:
            lim_1 = 1
        if lim_2 > dim_radio - 2:
            lim_2 = dim_radio - 2

        if lim_2 <= lim_1:
            message = (
                "Image shape or cropped selection too small for global search."
                + " After removal of the margins the search limits collide."
                + " The cropped size is %d\n" % (dim_radio)
            )
            raise ValueError(message)

        found_centers = []
        x_cor = lim_1
        while x_cor < lim_2:

            tmp_sigma = min((img_1.shape[1] - x_cor), (x_cor),) * self.sigma_fraction

            tmp_x = (np.arange(img_1.shape[1]) - x_cor) / tmp_sigma
            apodis = np.exp(-tmp_x * tmp_x / 2.0)

            x_cor_rel = x_cor - (img_1.shape[1] // 2)

            img_1_apodised = img_1 * apodis

            try:
                cor_position = CenterOfRotation.find_shift(
                    self,
                    img_1_apodised.astype(used_type),
                    img_2.astype(used_type),
                    low_pass=low_pass,
                    high_pass=high_pass,
                    roi_yxhw=roi_yxhw,
                )
            except ValueError as err:
                message = "ValueError from base class CenterOfRotation.find_shift in CenterOfRotationAdaptiveSearch.find_shift  : {err}".format(
                    err=err
                )
                self.logger.warning(message)
                x_cor = min(x_cor + x_cor * self.step_fraction, x_cor + (dim_radio - x_cor) * self.step_fraction)
                continue
            except:
                message = "Unexpected error from base class CenterOfRotation.find_shift in CenterOfRotationAdaptiveSearch.find_shift  : {err}".format(
                    err=err
                )
                self.logger.error(message)
                raise

            p_1 = cor_position * 2
            if cor_position < 0:
                p_2 = img_2.shape[1] + cor_position * 2
            else:
                p_2 = -img_2.shape[1] + cor_position * 2

            if abs(x_cor_rel - p_1 / 2) < abs(x_cor_rel - p_2 / 2):
                cor_position = p_1 / 2
            else:
                cor_position = p_2 / 2

            cor_in_img = img_1.shape[1] // 2 + cor_position
            tmp_sigma = min((img_1.shape[1] - cor_in_img), (cor_in_img),) * self.sigma_fraction

            M1 = int(round(cor_position + img_1.shape[1] // 2)) - int(round(tmp_sigma))
            M2 = int(round(cor_position + img_1.shape[1] // 2)) + int(round(tmp_sigma))

            piece_1 = img_filtered_1[:, M1:M2]
            piece_2 = img_filtered_2[:, img_1.shape[1] - M2 : img_1.shape[1] - M1]

            if piece_1.size and piece_2.size:
                piece_1 = piece_1 - piece_1.mean()
                piece_2 = piece_2 - piece_2.mean()
                energy = np.array(piece_1 * piece_1 + piece_2 * piece_2, "d").sum()
                diff_energy = np.array((piece_1 - piece_2) * (piece_1 - piece_2), "d").sum()
                cost = diff_energy / energy

                if not np.isnan(cost):
                    if tmp_sigma * 2 > abs(x_cor_rel - cor_position):
                        found_centers.append([cost, abs(x_cor_rel - cor_position), cor_position, energy])

            x_cor = min(x_cor + x_cor * self.step_fraction, x_cor + (dim_radio - x_cor) * self.step_fraction)

        if len(found_centers) == 0:
            message = "Unable to find any valid CoR candidate in {my_class}.find_shift ".format(
                my_class=self.__class__.__name__
            )
            raise ValueError(message)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Now build the neigborhood of the minimum as a list of five elements:
        # the minimum in the middle of the two before, and the two after

        filtered_found_centers = []
        for i in range(len(found_centers)):
            if i > 0:
                if abs(found_centers[i][2] - found_centers[i - 1][2]) < 0.5:
                    filtered_found_centers.append(found_centers[i])
                    continue
            if i + 1 < len(found_centers):
                if abs(found_centers[i][2] - found_centers[i + 1][2]) < 0.5:
                    filtered_found_centers.append(found_centers[i])
                    continue

        if len(filtered_found_centers):
            found_centers = filtered_found_centers

        min_choice = min(found_centers)
        index_min_choice = found_centers.index(min_choice)
        min_neighborood = [
            found_centers[i][2] if (i >= 0 and i < len(found_centers)) else math.nan
            for i in range(index_min_choice - 2, index_min_choice + 2 + 1)
        ]

        score_right = 0
        for i_pos in [3, 4]:
            if abs(min_neighborood[i_pos] - min_neighborood[2]) < 0.5:
                score_right += 1
            else:
                break

        score_left = 0
        for i_pos in [1, 0]:
            if abs(min_neighborood[i_pos] - min_neighborood[2]) < 0.5:
                score_left += 1
            else:
                break

        if score_left + score_right >= 2:
            validity_check_result = cor_result_validity["sound"]
        else:
            self.logger.warning(
                "Minimum mean difference and minimum std-dev of differences do not coincide. "
                + "This means that the validity of the found solution might be questionable."
            )
            validity_check_result = cor_result_validity["questionable"]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # An informative message in case one wish to look at how it has gone
        informative_message = " ".join(
            ["CenterOfRotationAdaptiveSearch found this neighborood of the optimal position:"]
            + [str(t) if not math.isnan(t) else "N.A." for t in min_neighborood]
        )
        self.logger.debug(informative_message)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # The return value is the optimum which had been placed in the middle of the neighborood
        cor_position = min_neighborood[2]

        if return_validity:
            return cor_position, validity_check_result
        else:
            return cor_position

    __call__ = find_shift
